import os
from functools import cached_property
from pathlib import Path

import yaml

from kubetorch.logger import get_logger


logger = get_logger(__name__)


ENV_MAPPINGS = {
    "username": "KT_USERNAME",
    "license_key": "KT_LICENSE_KEY",
    "namespace": "KT_NAMESPACE",
    "install_namespace": "KT_INSTALL_NAMESPACE",
    "install_url": "KT_INSTALL_URL",
    "stream_logs": "KT_STREAM_LOGS",
    "stream_metrics": "KT_STREAM_METRICS",
    "log_verbosity": "KT_LOG_VERBOSITY",
    "queue": "KT_QUEUE",
    "volumes": "KT_VOLUMES",
    "api_url": "KT_API_URL",
    "cluster_config": "KT_CLUSTER_CONFIG",
}

DEFAULT_INSTALL_NAMESPACE = "kubetorch"


class KubetorchConfig:
    CONFIG_FILE = Path("~/.kt/config.yaml")

    def __init__(self):
        self._api_url = None
        self._cluster_config = None
        self._install_namespace = None
        self._install_url = None
        self._license_key = None
        self._log_verbosity = None
        self._namespace = None
        self._queue = None
        self._stream_logs = None
        self._stream_metrics = None
        self._username = None
        self._volumes = None

    @cached_property
    def file_cache(self):
        return self._load_from_file()

    @cached_property
    def current_context(self):
        try:
            from kubetorch.servers.http.utils import is_running_in_kubernetes

            if is_running_in_kubernetes():
                try:
                    with open("/var/run/secrets/kubernetes.io/serviceaccount/namespace") as f:
                        return f.read().strip()
                except FileNotFoundError:
                    return "default"

            else:
                from kubernetes import config

                from kubetorch.utils import load_kubeconfig

                load_kubeconfig()
                _, active_context = config.list_kube_config_contexts()
                return active_context.get("context", {}).get("namespace", "default")

        except Exception:
            return "default"

    @property
    def username(self):
        """Username to use for Kubetorch deployments.

        Used for authentication and resource naming. Will be validated to ensure Kubernetes compatibility.
        """
        if not self._username:
            if self._get_env_var("username"):
                self._username = self._get_env_var("username")
            else:
                self._username = self.file_cache.get("username")
        return self._username

    @username.setter
    def username(self, value):
        """Set kubetorch username for current process."""
        from kubetorch.utils import validate_username

        validated = validate_username(value)
        if validated != value:
            logger.info(f"Username was validated and changed to {validated} to be Kubernetes-compatible.")
        self._username = validated

    @property
    def license_key(self):
        """License key for authentication and billing.

        Required for usage reporting and cluster authentication.
        Can be found in the `basic install guide <https://www.run.house/kubetorch/installation>`_.
        """
        if not self._license_key:
            if self._get_env_var("license_key"):
                self._license_key = self._get_env_var("license_key")
            else:
                self._license_key = self.file_cache.get("license_key")
        return self._license_key

    @license_key.setter
    def license_key(self, value: str):
        """Set kubetorch license key for current process."""
        self._license_key = value

    @property
    def queue(self):
        """Default queue name for scheduling services.

        Controls how cluster resources are allocated and prioritized for services.
        See `scheduling and queues <https://www.run.house/kubetorch/advanced-installation#scheduling-and-queues>`_ for more info.
        """
        if not self._queue:
            if self._get_env_var("queue"):
                self._queue = self._get_env_var("queue")
            else:
                self._queue = self.file_cache.get("queue")
        return self._queue

    @queue.setter
    def queue(self, value: str):
        self._queue = value

    @property
    def volumes(self):
        if not self._volumes:
            if self._get_env_var("volumes"):
                self._volumes = self._get_env_var("volumes")
            else:
                self._volumes = self.file_cache.get("volumes")
        return self._volumes

    @volumes.setter
    def volumes(self, values):
        if values is None or values == "None":
            self._volumes = None
        elif isinstance(values, str):
            # Handle comma-separated string
            self._volumes = [v.strip() for v in values.split(",") if v.strip()]
        elif isinstance(values, list):
            self._volumes = values
        else:
            raise ValueError("volumes must be a list of strings or comma-separated string")

    @property
    def api_url(self):
        if not self._api_url:
            if self._get_env_var("api_url"):
                self._api_url = self._get_env_var("api_url")
            else:
                self._api_url = self.file_cache.get("api_url")
        return self._api_url

    @api_url.setter
    def api_url(self, value: str):
        self._api_url = value

    @property
    def namespace(self):
        """Default Kubernetes namespace for Kubetorch deployments.

        All services will be deployed to this namespace unless overridden in the
        Compute resource constructor. If `install_namespace` is set, it will override this namespace.

        Priority:
            1. Explicit override
            2. Environment variable
            3. File cache
            4. In-cluster namespace or kubeconfig current context
        """
        if self.install_namespace and self.install_namespace != DEFAULT_INSTALL_NAMESPACE:
            self._namespace = self.install_namespace
        elif self._namespace is None:
            ns = self._get_env_var("namespace") or self.file_cache.get("namespace")
            self._namespace = ns or self.current_context
        return self._namespace

    @namespace.setter
    def namespace(self, value):
        """Set namespace for current process."""
        self._namespace = value

    @property
    def install_namespace(self):
        """Namespace for Kubetorch installation. Used for Kubetorch Cloud clients.

        Priority:
            1. Explicit override
            2. Environment variable
            3. File cache
            4. Default install namespace
        """
        if self._install_namespace is None:
            ns = self._get_env_var("install_namespace") or self.file_cache.get("install_namespace")
            self._install_namespace = ns or DEFAULT_INSTALL_NAMESPACE
        return self._install_namespace

    @install_namespace.setter
    def install_namespace(self, value):
        """Set installnamespace for current process."""
        self._install_namespace = value

    @property
    def install_url(self):
        """URL of the Kubetorch version to install.

        Used when installing Kubetorch in a Docker image or remote environment.
        Can be found in the `basic install guide <https://www.run.house/kubetorch/installation>`_.
        """
        if self._install_url is None:
            if self._get_env_var("install_url"):
                self._install_url = self._get_env_var("install_url")
            else:
                self._install_url = self.file_cache.get("install_url")
        return self._install_url

    @install_url.setter
    def install_url(self, value):
        """Set default kubetorch install url in current process."""
        self._install_url = value

    @property
    def log_verbosity(self):
        """Verbosity of logs streamed from a remote deployment.
        Log levels include ``debug``, ``info``, and ``critical``. Default is ``info``.

        Note:
            Only relevant when ``stream_logs`` is set to ``true``.
        """
        from kubetorch.utils import LogVerbosity

        default_verbosity = LogVerbosity.INFO.value

        if self._log_verbosity is None:
            verbosity_env_var = self._get_env_var("log_verbosity")
            if verbosity_env_var:
                try:
                    verbosity_env_var = LogVerbosity(verbosity_env_var).value
                except ValueError:
                    verbosity_env_var = default_verbosity

                self._log_verbosity = verbosity_env_var
            else:
                self._log_verbosity = self.file_cache.get("log_verbosity", default_verbosity)

        return self._log_verbosity

    @log_verbosity.setter
    def log_verbosity(self, value):
        """Set log verbosity."""
        from kubetorch.utils import LogVerbosity

        try:
            # In case we are unsetting log_verbosity, None is a valid value
            verbosity = LogVerbosity(value).value if value else None
        except ValueError:
            raise ValueError("Invalid log verbosity value. Must be one of: 'debug', 'info', 'critical'.")

        self._log_verbosity = verbosity

    @property
    def stream_logs(self):
        """Whether to stream logs for Kubetorch services.

        When enabled, logs from remote services are streamed back to your local environment
        in real-time. Verbosity of the streamed logs can be controlled with ``log_verbosity``.
        Default is ``True``

        When disabled, logs remain accessible in-cluster but are not streamed to the client.

        Note:
            Requires logging to be configured in the cluster (`logStreaming.enabled: true`` in the Helm chart)
        """
        if self._stream_logs is None:
            if self._get_env_var("stream_logs"):
                self._stream_logs = self._get_env_var("stream_logs").lower() == "true"
            else:
                self._stream_logs = self.file_cache.get("stream_logs", True)  # Default to True
        return self._stream_logs

    @stream_logs.setter
    def stream_logs(self, value):
        """Set log streaming for current process."""
        from kubetorch.serving.utils import check_loki_enabled

        bool_value = value

        if not isinstance(value, bool):
            if value is None:
                pass  # case we are unsetting stream_logs, so None is a valid value
            elif isinstance(value, str) and value.lower() in ["true", "false"]:
                bool_value = value.lower() == "true"
            else:
                raise ValueError("stream_logs must be a boolean value")
        if bool_value:
            # Check if the cluster has loki enabled
            if not check_loki_enabled():
                raise ValueError(
                    "Log streaming is not enabled in the cluster. Set `stream_logs` to False or "
                    "re-install the Kubetorch Helm chart with `logStreaming.enabled = true`"
                )
        self._stream_logs = bool_value

    @property
    def stream_metrics(self):
        """Whether to stream metrics during execution of Kubetorch services.

        When enabled, real-time CPU, memory, and GPU utilization metrics from remote Kubetorch services
        are streamed back to the local environment for live monitoring.
        Default is ``True``.

        When disabled, metrics are not collected.

        Note:
            Requires monitoring to be configured in the cluster (`metrics.enabled: true`` in the Helm chart)
        """
        if self._stream_metrics is None:
            if self._get_env_var("stream_metrics"):
                self._stream_metrics = self._get_env_var("stream_metrics").lower() == "true"
            else:
                self._stream_metrics = self.file_cache.get("stream_metrics", True)  # Default to True
        return self._stream_metrics

    @stream_metrics.setter
    def stream_metrics(self, value):
        """Set metrics streaming for current process."""
        from kubetorch.serving.utils import check_prometheus_enabled

        bool_value = value

        if not isinstance(value, bool):
            if value is None:
                pass  # case we are unsetting stream_metrics, so None is a valid value
            elif isinstance(value, str) and value.lower() in ["true", "false"]:
                bool_value = value.lower() == "true"
            else:
                raise ValueError("stream_metrics must be a boolean value")
        if bool_value:
            # Check if the cluster has prometheus enabled
            if not check_prometheus_enabled():
                raise ValueError(
                    "Metrics is not enabled in the cluster. Set `stream_metrics` to False or "
                    "re-install the Kubetorch Helm chart with `metrics.enabled = true`"
                )
        self._stream_metrics = bool_value

    @property
    def cluster_config(self):
        """Cluster Config.
        Default is ``{}``.
        """
        from kubetorch.utils import string_to_dict

        config = self._cluster_config
        if self._cluster_config is None:
            config = string_to_dict(self._get_env_var("cluster_config") or "")
            if not config:
                config = string_to_dict(self.file_cache.get("cluster_config", "{}"))
            self._cluster_config = config
        return config

    @cluster_config.setter
    def cluster_config(self, value):
        """Set Cluster Config."""
        from kubetorch.utils import string_to_dict

        new_value = value
        if not isinstance(new_value, dict):
            if isinstance(new_value, str):
                new_value = string_to_dict(new_value)
            else:
                new_value = {}  # Default to empty dict
        self._cluster_config = new_value

    def __iter__(self):
        for key in ENV_MAPPINGS:
            value = getattr(self, key)
            if value == "None":
                value = None
            yield key, value

    def set(self, key, value):
        if key not in ENV_MAPPINGS:
            raise ValueError(f"Unknown config key: {key}")
        setattr(self, key, value)
        # if key is 'username' and value is None (unsetting username), we'll get the cached username,and not the
        # new value
        new_value = value if value is None else getattr(self, key)
        return new_value

    def get(self, key):
        if key not in ENV_MAPPINGS:
            raise ValueError(f"Unknown config key: {key}")
        return getattr(self, key)

    def write(self, values: dict = None):
        """Write out config to local ``~/.kt/config.yaml``, to be used globally.

        Args:
            values (optional): Dict of key-value pairs to write/update in the filesystem.
                If provided, only these keys will be updated. None values will remove the key from the file.
        """
        # Ensure directory exists
        self.CONFIG_FILE.expanduser().parent.mkdir(parents=True, exist_ok=True)

        if values:
            values_to_write = self._load_from_file()
            for k, v in values.items():
                if k not in ENV_MAPPINGS:
                    raise ValueError(f"Unknown config key: {k}")
                if v is None:
                    values_to_write.pop(k, None)
                else:
                    values_to_write[k] = str(v) if isinstance(v, dict) else v
        else:
            values_to_write = {k: str(v) if isinstance(v, dict) else v for k, v in dict(self).items() if v is not None}

        # Write to file
        with self.CONFIG_FILE.expanduser().open("w") as stream:
            yaml.safe_dump(values_to_write, stream)

        # Invalidate file cache so it reloads on next access
        if "file_cache" in self.__dict__:
            del self.__dict__["file_cache"]

    def _get_env_var(self, key):
        return os.getenv(ENV_MAPPINGS[key])

    def _get_config_env_vars(self):
        """Get config values as environment variables with proper KT_ prefixes."""
        env_vars = {}
        for key, value in dict(self).items():
            if value is not None and key in ENV_MAPPINGS:
                env_vars[ENV_MAPPINGS[key]] = value
        return env_vars

    def _load_from_file(self):
        if self.CONFIG_FILE.expanduser().exists():
            with open(self.CONFIG_FILE.expanduser(), "r") as stream:
                return yaml.safe_load(stream) or {}
        return {}



from dataclasses import asdict, dataclass

from kubetorch.logger import get_logger

logger = get_logger(__name__)


class AutoScalingError(Exception):
    pass


@dataclass
class AutoscalingConfig:
    # The concurrent requests or requests per second threshold that triggers scaling
    target: int = None

    # The time window used to calculate the average of metrics for scaling decisions
    window: str = None

    # The metric type to base scaling decisions on:
    # - concurrency: number of simultaneous requests
    # - rps: requests per second
    # - cpu: CPU utilization (requires HPA class)
    # - memory: Memory utilization (requires HPA class)
    metric: str = None

    # The percentage of the target value at which to start scaling.
    # E.g., if target=100 and target_utilization=70, scaling occurs at 70 requests
    target_utilization: int = None

    # Minimum number of replicas. 0 allows scaling to zero when idle
    min_scale: int = None

    # Maximum number of replicas the service can scale up to
    max_scale: int = None

    # Initial number of pods launched by the service
    initial_scale: int = None

    # Maximum concurrent requests per pod (containerConcurrency).
    # If not set, pods accept unlimited concurrent requests.
    concurrency: int = None

    # Time to keep the last pod before scaling to zero (e.g., "30s", "1m5s")
    scale_to_zero_pod_retention_period: str = None

    # Delay before scaling down (e.g., "15m"). Only for KPA autoscaler.
    scale_down_delay: str = None

    # Autoscaler class: "kpa.autoscaling.knative.dev" or "hpa.autoscaling.knative.dev"
    autoscaler_class: str = None

    # Progress deadline for deployment (e.g., "10m"). Time to wait for deployment to be ready.
    progress_deadline: str = None

    def __init__(self, **kwargs):
        """Support additional kwargs for autoscaling annotations"""
        for field in self.__annotations__:
            setattr(self, field, kwargs.pop(field, getattr(self, field, None)))

        # set additional kwargs as annotations
        self.extra_annotations = {f"autoscaling.knative.dev/{k}": str(v) for k, v in kwargs.items()}

        self._validate()

    def _validate(self):
        """Validation logic moved to separate method"""
        if self.min_scale is not None and self.max_scale is not None and self.min_scale > self.max_scale:
            raise AutoScalingError("min_scale cannot be greater than max_scale")
        if self.window is not None and not self.window.endswith(("s", "m", "h")):
            raise AutoScalingError("window must end with s, m, or h")
        if self.target_utilization is not None and (self.target_utilization <= 0 or self.target_utilization > 100):
            raise AutoScalingError("target_utilization must be between 1 and 100")
        if self.scale_to_zero_pod_retention_period is not None:
            # Validate time format (e.g., "30s", "1m5s", "2h")
            import re

            if not re.match(r"^\d+[smh](\d+[smh])*$", self.scale_to_zero_pod_retention_period):
                raise AutoScalingError(
                    "scale_to_zero_pod_retention_period must be a valid duration (e.g., '30s', '1m5s')"
                )
        if self.scale_down_delay is not None:
            # Validate time format
            import re

            if not re.match(r"^\d+[smh](\d+[smh])*$", self.scale_down_delay):
                raise AutoScalingError("scale_down_delay must be a valid duration (e.g., '15m', '1h')")
        if self.autoscaler_class is not None and self.autoscaler_class not in [
            "kpa.autoscaling.knative.dev",
            "hpa.autoscaling.knative.dev",
        ]:
            raise AutoScalingError(
                "autoscaler_class must be 'kpa.autoscaling.knative.dev' or 'hpa.autoscaling.knative.dev'"
            )
        if self.progress_deadline is not None:
            # Validate time format
            import re

            if not re.match(r"^\d+[smh](\d+[smh])*$", self.progress_deadline):
                raise AutoScalingError("progress_deadline must be a valid duration (e.g., '10m', '600s')")

    def __post_init__(self):
        """Call the same validation for dataclass initialization"""
        self._validate()

    def dict(self):
        return asdict(self)

    def convert_to_annotations(self) -> dict:
        """Convert config to a dictionary of annotations for Knative"""
        annotations = {}

        # Set autoscaler class if specified, otherwise use default KPA
        if self.autoscaler_class is not None:
            annotations["autoscaling.knative.dev/class"] = self.autoscaler_class
        else:
            annotations["autoscaling.knative.dev/class"] = "kpa.autoscaling.knative.dev"

        # Only set annotations for values that were explicitly provided
        if self.target is not None:
            annotations["autoscaling.knative.dev/target"] = str(self.target)

        if self.min_scale is not None:
            annotations["autoscaling.knative.dev/min-scale"] = str(self.min_scale)

        if self.max_scale is not None:
            annotations["autoscaling.knative.dev/max-scale"] = str(self.max_scale)

        if self.window is not None:
            annotations["autoscaling.knative.dev/window"] = self.window

        if self.metric is not None:
            annotations["autoscaling.knative.dev/metric"] = self.metric

        if self.target_utilization is not None:
            annotations["autoscaling.knative.dev/target-utilization-percentage"] = str(self.target_utilization)

        if self.initial_scale is not None:
            annotations["autoscaling.knative.dev/initial-scale"] = str(self.initial_scale)

        if self.scale_to_zero_pod_retention_period is not None:
            annotations[
                "autoscaling.knative.dev/scale-to-zero-pod-retention-period"
            ] = self.scale_to_zero_pod_retention_period

        if self.scale_down_delay is not None:
            annotations["autoscaling.knative.dev/scale-down-delay"] = self.scale_down_delay

        # Add any extra annotations from the config
        if hasattr(self, "extra_annotations"):
            annotations.update(self.extra_annotations)

        return annotations



import random
import socket
import threading
import time

import websocket


class WebSocketRsyncTunnel:
    def __init__(self, local_port: int, ws_url: str):
        self.requested_port = local_port
        self.local_port = None  # Will be set in __enter__
        self.ws_url = ws_url
        self.running = False
        self.server_socket = None

    def __enter__(self):
        self.running = True

        # Add randomization to reduce concurrent collision probability
        # Try multiple times with different socket instances
        max_attempts = 20
        port_range = 100  # Much wider range to avoid collisions

        for attempt in range(max_attempts):
            # Add random offset to spread out concurrent requests
            random_offset = random.randint(0, 50)
            start_port = self.requested_port + random_offset

            # Create a new socket for each attempt
            server_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            server_socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)

            # Try ports in this range
            for i in range(port_range):
                port = start_port + i
                try:
                    server_socket.bind(("127.0.0.1", port))
                    server_socket.listen(5)
                    # Success! Save the socket and port
                    self.server_socket = server_socket
                    self.local_port = port
                    break
                except OSError as e:
                    if e.errno == 98:  # Address already in use
                        continue
                    else:
                        server_socket.close()
                        raise
            else:
                # No ports available in this range, close socket and try again
                server_socket.close()
                # Add small random delay to reduce simultaneous collision probability
                time.sleep(random.uniform(0.01, 0.05))
                continue

            # Websocket server successfully bound and is listening
            break
        else:
            raise RuntimeError(
                f"Could not find available port after {max_attempts} attempts starting from {self.requested_port}"
            )

        threading.Thread(target=self._accept_loop, daemon=True).start()

        # Wait for ready
        for _ in range(50):
            try:
                with socket.socket() as s:
                    s.settimeout(0.1)
                    if s.connect_ex(("127.0.0.1", self.local_port)) == 0:
                        return self
            except:
                pass
            time.sleep(0.1)
        raise RuntimeError("Tunnel failed to start")

    def __exit__(self, *args):
        self.running = False
        if self.server_socket:
            try:
                self.server_socket.close()
            except:
                pass  # Already closed

    def _accept_loop(self):
        while self.running:
            try:
                client_sock, _ = self.server_socket.accept()
                threading.Thread(target=self._handle_client, args=(client_sock,), daemon=True).start()
            except:
                break

    def _handle_client(self, client_sock):
        ws = None
        try:
            ws = websocket.create_connection(self.ws_url)

            def tcp_to_ws():
                while self.running:
                    try:
                        data = client_sock.recv(65536)
                        if not data:
                            break
                        ws.send_binary(data)
                    except:
                        break

            def ws_to_tcp():
                while self.running:
                    try:
                        data = ws.recv()
                        if isinstance(data, bytes):
                            client_sock.send(data)
                    except:
                        break

            t1 = threading.Thread(target=tcp_to_ws, daemon=True)
            t2 = threading.Thread(target=ws_to_tcp, daemon=True)
            t1.start()
            t2.start()
            t1.join()
            t2.join()

        except Exception as e:
            print(f"WebSocket connection error: {e}")

        finally:
            # close both connections
            for conn in [client_sock, ws]:
                if conn:
                    try:
                        conn.close()
                    except:
                        pass



import base64
import importlib
import inspect
import os
import signal
import subprocess
import sys
import textwrap
import time
from datetime import datetime
from pathlib import Path
from typing import List
from urllib.parse import urlparse

import httpx
from kubernetes.client.rest import ApiException
from rich.syntax import Syntax

from kubetorch.servers.http.utils import is_running_in_kubernetes

from .cli_utils import (
    create_table_for_output,
    default_typer_values,
    get_deployment_mode,
    get_ingress_host,
    get_last_updated,
    get_logs_from_loki,
    is_ingress_vpc_only,
    load_ingress,
    load_kubetorch_volumes_for_service,
    port_forward_to_pod,
    SecretAction,
    service_name_argument,
    validate_config_key,
    validate_pods_exist,
    validate_provided_pod,
    VolumeAction,
)

from .utils import initialize_k8s_clients

try:
    import typer

    from rich.console import Console
    from rich.panel import Panel
    from rich.table import Table
except ImportError:
    raise ImportError("Please install the required CLI dependencies: `pip install 'kubetorch[client] @ <install_url>'`")


import kubetorch.serving.constants as serving_constants

from kubetorch import globals
from kubetorch.config import ENV_MAPPINGS
from kubetorch.servers.http.utils import DEFAULT_DEBUG_PORT

from .constants import BULLET_UNICODE, KT_MOUNT_FOLDER

try:
    from .internal.cli import register_internal_commands

    _INTERNAL_COMMANDS_AVAILABLE = True
except ImportError:
    _INTERNAL_COMMANDS_AVAILABLE = False

from .logger import get_logger

app = typer.Typer(add_completion=False)
console = Console()

# Register internal CLI commands if available
if _INTERNAL_COMMANDS_AVAILABLE:
    register_internal_commands(app)


logger = get_logger(__name__)


@app.command("check")
def kt_check(
    name: str = service_name_argument(help="Service name"),
    namespace: str = typer.Option(
        globals.config.namespace,
        "-n",
        "--namespace",
    ),
):
    """
    Run a comprehensive health check for a deployed service.

    Checks:

    - Deployment pod comes up and becomes ready (if not scaled to zero)

    - Rsync has succeeded

    - Service is marked as ready and service pod(s) are ready to serve traffic

    - GPU support configured (if applicable)

    - Log streaming configuration (if applicable)

    If a step fails, will dump ``kubectl describe`` and pod logs for relevant pods.
    """
    core_api, custom_api, apps_v1_api = initialize_k8s_clients()

    def dump_pod_debug(pod_name):
        try:
            describe_proc = subprocess.run(
                ["kubectl", "describe", "pod", pod_name, "-n", namespace],
                check=False,
                capture_output=True,
                text=True,
            )
            describe_output = describe_proc.stdout or describe_proc.stderr or "<no output>"

            logs_proc = subprocess.run(
                ["kubectl", "logs", pod_name, "-n", namespace, "-c", "kubetorch"],
                check=False,
                capture_output=True,
                text=True,
            )
            logs_output = logs_proc.stdout or logs_proc.stderr or "<no output>"

            console.print(
                Panel(
                    describe_output,
                    title=f"POD DESCRIPTION ({pod_name})",
                    border_style="yellow",
                    expand=False,
                )
            )
            console.print(
                Panel(
                    logs_output,
                    title=f"POD LOGS ({pod_name})",
                    border_style="yellow",
                    expand=False,
                )
            )
        except Exception as e:
            console.print(f"[red]Failed to dump pod info: {e}[/red]")

    def fail(msg, pod_names=None):
        console.print(f"[red]{msg}[/red]")
        if pod_names:
            for pod_name in pod_names:
                dump_pod_debug(pod_name)
        raise typer.Exit(1)

    try:
        # Validate service exists and get deployment mode
        name, deployment_mode = get_deployment_mode(name, namespace, custom_api, apps_v1_api)

        console.print(f"[bold blue]Checking {deployment_mode} service...[/bold blue]")

        # 1. Deployment pod check
        console.print("[bold blue]Checking deployment pod...[/bold blue]")
        deploy_pods = validate_pods_exist(name, namespace, core_api)

        if not deploy_pods:
            if deployment_mode == "knative":
                try:
                    # Check if the Knative service is marked as ready (e.g. scaled to zero)
                    service = custom_api.get_namespaced_custom_object(
                        group="serving.knative.dev",
                        version="v1",
                        namespace=namespace,
                        plural="services",
                        name=name,
                    )
                    conditions = service.get("status", {}).get("conditions", [])
                    ready = any(c.get("type") == "Ready" and c.get("status") == "True" for c in conditions)
                    if ready:
                        console.print(
                            f"[yellow]No deployment pods found. Service [bold]{name}[/bold] is scaled to zero but marked as 'READY'. "
                            "It will scale up on demand.[/yellow]"
                        )
                        return
                    else:
                        fail("Deployment pod not found and service is not READY.")

                except Exception as e:
                    fail(f"Failed to check Knative service status: {e}")
            else:
                fail("No Deployment pods found.")

        deploy_pod = next(
            (p for p in deploy_pods if p.status.phase == "Running" and not p.metadata.deletion_timestamp),
            None,
        )
        if not deploy_pod:
            fail(
                "No deployment pod in 'Running' state found.",
                [p.metadata.name for p in deploy_pods],
            )

        deploy_pod_name = deploy_pod.metadata.name
        if deploy_pod.status.phase != "Running":
            fail(
                f"Deployment pod not running (status: {deploy_pod.status.phase})",
                [deploy_pod_name],
            )

        # 2. Rsync check
        console.print("[bold blue]Checking rsync...[/bold blue]")
        current_working_dir = "."
        check_cmd = [
            "kubectl",
            "exec",
            deploy_pod_name,
            "-n",
            namespace,
            "--",
            "ls",
            "-l",
            current_working_dir,
        ]
        try:
            result = subprocess.run(check_cmd, capture_output=True, text=True, check=True)
            lines = result.stdout.splitlines()
            entries = [line for line in lines if not line.startswith("total")]
            if not entries:
                fail("Rsync directory exists but is empty.", [deploy_pod_name])
        except subprocess.CalledProcessError as e:
            fail(
                f"Rsync directory check failed: {e.stderr or e.stdout}",
                [deploy_pod_name],
            )

        # 3. Service call check
        console.print("[bold blue]Checking service call...[/bold blue]")
        try:
            with port_forward_to_pod(
                pod_name=deploy_pod_name,
                namespace=namespace,
                local_port=32300,
                remote_port=32300,
            ) as local_port:
                url = f"http://localhost:{local_port}/health"
                resp = httpx.get(url, timeout=10)
                if not resp.is_success:
                    fail(
                        f"Service call failed: {resp.status_code} {resp.text}",
                        [deploy_pod_name],
                    )
        except Exception as e:
            fail(f"Service call check failed: {e}", [deploy_pod_name])

        # 5. GPU + autoscaler test (if GPU requested)
        gpu_requested = any(
            c.resources.limits and "nvidia.com/gpu" in c.resources.limits for c in deploy_pod.spec.containers
        )
        if gpu_requested:
            gpus_configured = False
            console.print("[bold blue]Checking GPU plugin support...[/bold blue]")
            nodes = core_api.list_node().items
            for node in nodes:
                gpus = node.status.capacity.get("nvidia.com/gpu")
                if gpus and int(gpus) > 0:
                    gpus_configured = True
                    break

            if not gpus_configured:
                console.print(
                    "[yellow]No GPU nodes currently configured on the cluster, is autoscaling configured?[/yellow]"
                )

            dcgm_exporter = True
            dcgm_namespace = globals.config.install_namespace

            pods = core_api.list_namespaced_pod(
                namespace=dcgm_namespace,
                label_selector="app.kubernetes.io/name=dcgm-exporter",
            ).items
            if not pods:
                dcgm_exporter = False

            if not dcgm_exporter:
                console.print(f"[yellow]DCGM exporter not found in namespace {dcgm_namespace}[/yellow]")

        # 6. Check logs
        if globals.config.stream_logs:
            try:
                streaming_enabled = core_api.read_namespaced_service(
                    name=serving_constants.LOKI_GATEWAY_SERVICE_NAME,
                    namespace=globals.config.install_namespace,
                )
            except ApiException:
                streaming_enabled = False

            if streaming_enabled:

                console.print("[bold blue]Checking log streaming...[/bold blue]")
                query = f'{{k8s_pod_name="{deploy_pod_name}", k8s_container_name="kubetorch"}}'
                try:
                    logs = get_logs_from_loki(query=query, print_pod_name=False, timeout=5.0)
                    if logs is None:
                        fail("No logs found for service", [deploy_pod_name])

                except Exception as e:
                    fail(f"Logs check failed: {e}", [deploy_pod_name])

        console.print("[bold green]✓ All service checks passed[/bold green]")

    except typer.Exit:
        # Just re-raise, don't print
        raise


@app.command("config")
def kt_config(
    action: str = typer.Argument(default="", help="Action to perform (set, unset, get, list)"),
    key: str = typer.Argument(None, help="Config key (e.g., 'username')", callback=validate_config_key),
    value: str = typer.Argument(None, help="Value to set"),
):
    """Manage Kubetorch configuration settings.

    Examples:

    .. code-block:: bash

        $ kt config set username johndoe

        $ kt config set volumes "volume_name_one, volume_name_two"

        $ kt config set volumes volume_name_one

        $ kt config unset username

        $ kt config get username

        $ kt config list
    """
    from kubetorch import config

    if action == "set":
        if not key or not value:
            console.print("[red]Both key and value are required for 'set'[/red]")
            raise typer.Exit(1)

        try:
            value = config.set(key, value)  # validate value
            config.write({key: value})
            console.print(f"[green]{key} set to:[/green] [blue]{value}[/blue]")
        except ValueError as e:
            console.print(f"[red]Error setting {key}:[/red] {str(e)}")
            raise typer.Exit(1)

    elif action == "unset":
        if not key:
            console.print("[red]Key is required for 'unset'[/red]")
            raise typer.Exit(1)

        try:
            config.set(key, None)
            config.write({key: None})
            console.print(f"[green]{key.capitalize()} unset[/green]")
        except ValueError as e:
            console.print(f"[red]Error unsetting {key}:[/red] {str(e)}")
            raise typer.Exit(1)

    elif action == "get":
        if not key:
            # Error panel
            console.print("[red]Key is required for 'get'[/red]")
            raise typer.Exit(1)

        if key in ENV_MAPPINGS:
            value = config.get(key)
            if value:
                console.print(f"[blue]{value}[/blue]")
            else:
                console.print(f"[yellow]{key.capitalize()} not set[/yellow]")
        else:
            console.print(f"[red]Unknown config key:[/red] [bold]{key}[/bold]")
            raise typer.Exit(1)

    elif action == "list" or not action:
        console.print(dict(config))

    else:
        console.print(f"[red]Unknown action:[/red] [bold]{action}[/bold]")
        console.print("\nValid actions are: set, get, list")
        raise typer.Exit(1)


@app.command("debug")
def kt_debug(
    pod: str = typer.Argument(..., help="Pod name"),
    namespace: str = typer.Option(
        globals.config.namespace,
        "-n",
        "--namespace",
    ),
    port: int = typer.Option(DEFAULT_DEBUG_PORT, help="Debug port used for remote debug server"),
):
    """Start an interactive debugging session on the pod, which will connect to the debug server inside the service.
    Before running this command, you must call a method on the service with pdb=True or add a
    kt.deep_breakpoint() call into your code to enable debugging.
    """
    import webbrowser

    if is_running_in_kubernetes():
        console.print(
            "[red]Debugging is not supported when running inside Kubernetes. Please run this command locally.[/red]"
        )
        raise typer.Exit(1)

    # Use the base path of web-pdb server as health endpoint because we're port-forwarding straight into the pod
    with port_forward_to_pod(
        namespace=namespace,
        pod_name=pod,
        local_port=port,
        remote_port=port,
        health_endpoint="/",
    ):
        debug_ui_url = f"http://localhost:{port}"
        console.print(f"Opening debug UI at [blue]{debug_ui_url}[/blue]")
        webbrowser.open(debug_ui_url)
        # Wait for the user to finish debugging
        console.print("[yellow]Press Ctrl+C to stop the debugging session and close the UI.[/yellow]")
        # Wait for a Ctrl+C to exit the debug session
        try:
            while True:
                time.sleep(1)
        except KeyboardInterrupt:
            console.print("\n[yellow]Debugging session ended.[/yellow]")
            raise typer.Exit(0)


@app.command("deploy")
def kt_deploy(
    target: str = typer.Argument(
        ...,
        help="Python module or file to deploy, optionally followed by a "
        "single function or class to deploy. e.e. `my_module:my_cls`, or "
        "`my_file.py`.",
    ),
):
    """Deploy a Python file or module to Kubetorch. This will deploy all functions and modules decorated with
    @kt.compute in the file or module."""
    from kubetorch.resources.compute.utils import _collect_modules

    os.environ["KT_CLI_DEPLOY_MODE"] = "1"
    to_deploy, target_fn_or_class = _collect_modules(target)

    if not target_fn_or_class:
        console.print(f"Found the following functions and classes to deploy in {target}:")
        for module in to_deploy:
            console.print(f"{BULLET_UNICODE} {module.name}")

    import asyncio

    async def deploy_all_async():
        tasks = []
        for module in to_deploy:
            console.print(f"Deploying {module.name}...")
            tasks.append(module.deploy_async())

        try:
            await asyncio.gather(*tasks)
            for module in to_deploy:
                console.print(f"Successfully deployed {module.name}.")
        except Exception as e:
            console.print(f"Failed to deploy one or more modules: {e}")
            raise e

    asyncio.run(deploy_all_async())

    if not target_fn_or_class:
        console.print(f"Successfully deployed functions and modules from {target}.")


@app.command("describe")
def kt_describe(
    name: str = service_name_argument(help="Service name"),
    namespace: str = typer.Option(
        globals.config.namespace,
        "-n",
        "--namespace",
    ),
):
    """
    Show basic info for calling the service depending on whether an ingress is configured.
    """

    core_api, custom_api, apps_v1_api = initialize_k8s_clients()

    endpoint_placeholder = "METHOD_OR_CLS_NAME"
    args_placeholder = []

    try:
        name, deployment_mode = get_deployment_mode(name, namespace, custom_api, apps_v1_api)
    except ApiException:
        console.print(f"[red] Failed to load service '{name}' in namespace '{namespace}'[/red]")
        raise typer.Exit(1)

    try:
        console.print()
        base_url = globals.config.api_url

        ingress = load_ingress()
        host = get_ingress_host(ingress) if ingress else f"{name}.{namespace}.svc.cluster.local"

        if not base_url:
            if not ingress:
                console.print("[yellow]No ingress found. Service is only accessible from inside the cluster.[/yellow]")
                base_url = f"http://{name}.{namespace}.svc.cluster.local"
            else:
                lb_ing = (
                    ingress.status.load_balancer.ingress[0]
                    if (ingress.status and ingress.status.load_balancer and ingress.status.load_balancer.ingress)
                    else None
                )

                address = lb_ing.hostname or lb_ing.ip if lb_ing else None
                if address:
                    base_url = f"http://{address}"
                else:
                    console.print("[yellow]Ingress found but no address, falling back to cluster-local.[/yellow]")
                    base_url = f"http://{name}.{namespace}.svc.cluster.local"
        else:
            parsed = urlparse(base_url)
            if not parsed.scheme:
                base_url = f"http://{base_url}"

        if ingress:
            console.print(f"[bold]Host:[/bold] [green]{name}[/green]")

            vpc_only = is_ingress_vpc_only(ingress.metadata.annotations)
            if vpc_only:
                console.print()
                console.print("[yellow]Note: This is a VPC-only ingress (internal access only)[/yellow]")

        console.print()

        if ingress:
            console.print("[bold]Calling the service using an ingress:[/bold]\n")
            # With ingress, use the full path structure
            service_path = f"/{namespace}/{name}/{endpoint_placeholder}"
        else:
            console.print("[bold]Calling the service from inside the cluster:[/bold]\n")
            service_path = f"/{endpoint_placeholder}"

        curl_code = textwrap.dedent(
            f"""\
            curl -X POST \\
              -H "Content-Type: application/json" \\
              -d '{{"args": {args_placeholder}, "kwargs": {{}}}}' \\
              {base_url}{service_path}
        """
        )
        # Only add Host header if we have ingress
        if ingress:
            curl_code = curl_code.replace(
                '-H "Content-Type: application/json"',
                f'-H "Host: {host}" \\\n  -H "Content-Type: application/json"',
            )

        console.print(Panel(Syntax(curl_code, "bash"), title="With Curl", border_style="green"))
        console.print()

        python_code = textwrap.dedent(
            f"""\
            import requests

            url = "{base_url}{service_path}"
            headers = {{
                "Content-Type": "application/json"
            }}
            data = {{
                "args": {args_placeholder},
                "kwargs": {{}}
            }}

            response = requests.post(url, headers=headers, json=data)
            print(response.json())
        """
        )
        if ingress:
            python_code = python_code.replace(
                '"Content-Type": "application/json"',
                f'"Host": "{host}",\n    "Content-Type": "application/json"',
            )
        console.print(
            Panel(
                Syntax(python_code, "python"),
                title="With Python",
                border_style="green",
            )
        )
    except Exception as e:
        console.print(
            f"[red]Failed to describe service {name} in namespace {namespace}: {e}[/red]",
        )
        raise typer.Exit(1)


@app.command("list")
def kt_list(
    namespace: str = typer.Option(
        globals.config.namespace,
        "-n",
        "--namespace",
    ),
    sort_by_updated: bool = typer.Option(False, "-s", "--sort", help="Sort by last update time"),
    tag: str = typer.Option(
        None,
        "-t",
        "--tag",
        help="Service tag or prefix (ex: 'myusername', 'some-git-branch').",
    ),
):
    """List all Kubetorch services.

    Examples:

    .. code-block:: bash

      $ kt list

      $ kt list -t dev-branch
    """
    core_api, custom_api, _ = initialize_k8s_clients()

    # Import here to avoid circular imports
    from kubetorch.serving.service_manager import BaseServiceManager

    try:
        # Use unified service discovery
        unified_services = BaseServiceManager.discover_services_static(namespace=namespace, name_filter=tag)

        if not unified_services:
            console.print(f"[yellow]No services found in {namespace} namespace[/yellow]")
            return

        # Optional second-level tag filtering
        if tag:
            unified_services = [
                svc
                for svc in unified_services
                if tag in svc["name"]
                or tag in " ".join(str(v) for v in svc["resource"].get("metadata", {}).get("labels", {}).values())
            ]
            if not unified_services:
                console.print(f"[yellow]No services found in {namespace} namespace[/yellow]")
                return

        if sort_by_updated:

            def get_update_time(svc):
                # If not a ksvc, use creation timestamp as proxy for update time
                return (
                    get_last_updated(svc["resource"]) if svc["template_type"] == "ksvc" else svc["creation_timestamp"]
                )

            unified_services.sort(key=get_update_time, reverse=True)

        # Get pod maps
        pod_map = {
            svc["name"]: BaseServiceManager.get_pods_for_service_static(svc["name"], namespace, core_api)
            for svc in unified_services
        }

        # Create table
        table_columns = [
            ("SERVICE", "cyan"),
            ("TYPE", "magenta"),
            ("STATUS", "green"),
            ("# OF PODS", "yellow"),
            ("POD NAMES", "red"),
            ("VOLUMES", "blue"),
            ("LAST STATUS CHANGE", "yellow"),
            ("TTL", "yellow"),
            ("CREATOR", "yellow"),
            ("QUEUE", "yellow"),
            ("CPUs", "yellow"),
            ("MEMORY", "yellow"),
            ("GPUs", "yellow"),
        ]
        table = create_table_for_output(
            columns=table_columns,
            no_wrap_columns_names=["SERVICE"],
            header_style={"bold": False},
        )

        for svc in unified_services:
            name = svc["name"]
            kind = svc["template_type"]
            res = svc["resource"]
            meta = res.get("metadata", {})
            labels = meta.get("labels", {})
            annotations = meta.get("annotations", {})
            status_data = res.get("status", {})

            # Get pods
            pods = pod_map.get(name, [])

            creation_ts = meta.get("creationTimestamp", None)
            timestamp = (
                datetime.fromisoformat(creation_ts.replace("Z", "+00:00")).strftime("%Y-%m-%d %H:%M:%S")
                if creation_ts
                else "Unknown"
            )
            ttl = annotations.get(serving_constants.INACTIVITY_TTL_ANNOTATION, "None")
            creator = labels.get(serving_constants.KT_USERNAME_LABEL, "—")

            volumes_display = load_kubetorch_volumes_for_service(namespace, name, core_api)

            # Get resources from revision
            cpu = memory = gpu = None
            if kind == "ksvc":
                cond = status_data.get("conditions", [{}])[0]
                status = cond.get("status")
                display_status = {
                    "True": "[green]Ready[/green]",
                    "Unknown": "[yellow]Creating[/yellow]",
                }.get(status, "[red]Failed[/red]")
                rev_name = status_data.get("latestCreatedRevisionName")
                if rev_name:
                    try:
                        rev = custom_api.get_namespaced_custom_object(
                            group="serving.knative.dev",
                            version="v1",
                            namespace=namespace,
                            plural="revisions",
                            name=rev_name,
                        )
                        container = rev["spec"]["containers"][0]
                        reqs = container.get("resources", {}).get("requests", {})
                        cpu = reqs.get("cpu")
                        memory = reqs.get("memory")
                        gpu = reqs.get("nvidia.com/gpu") or reqs.get("gpu")
                    except Exception as e:
                        logger.warning(f"Could not get revision for {name}: {e}")
            else:
                # Process Deployment - now using consistent dict access
                ready = res.get("status", {}).get("readyReplicas", 0) or 0
                desired = res.get("spec", {}).get("replicas", 0) or 0
                if kind == "raycluster":
                    state = status_data.get("state", "").lower()
                    conditions = {c["type"]: c["status"] for c in status_data.get("conditions", [])}
                    if (
                        state == "ready"
                        and conditions.get("HeadPodReady") == "True"
                        and conditions.get("RayClusterProvisioned") == "True"
                    ):
                        display_status = "[green]Ready[/green]"
                    elif state in ("creating", "upscaling", "restarting", "updating"):
                        display_status = "[yellow]Scaling[/yellow]"
                    else:
                        display_status = "[red]Failed[/red]"
                else:
                    display_status = (
                        "[green]Ready[/green]"
                        if ready == desired and desired > 0
                        else "[yellow]Scaling[/yellow]"
                        if ready < desired
                        else "[red]Failed[/red]"
                    )
                try:
                    container = res.get("spec", {}).get("template", {}).get("spec", {}).get("containers", [{}])[0]
                    reqs = container.get("resources", {}).get("requests", {})
                    cpu = reqs.get("cpu")
                    memory = reqs.get("memory")
                    gpu = reqs.get("nvidia.com/gpu") or reqs.get("gpu")
                except Exception as e:
                    logger.warning(f"Failed to get resources for {name} in namespace {namespace}: {e}")

            # Common pod processing
            pod_lines = []
            queue = "—"
            for pod in pods:
                pod_status = pod.status.phase
                ready = all(c.ready for c in (pod.status.container_statuses or []))
                if ready and pod_status == "Running":
                    color = "green"
                elif "Creating" in display_status or "Scaling" in display_status:
                    color = "yellow"
                else:
                    color = "red"
                pod_lines.append(f"[{color}]{pod.metadata.name}[/{color}]")
                queue = pod.metadata.labels.get(serving_constants.KAI_SCHEDULER_LABEL, queue)

                # Update service status if pod is pending
                if pod_status == "Pending":
                    display_status = "[yellow]Pending[/yellow]"

            table.add_row(
                name,
                f"[magenta]{kind}[/magenta]",
                display_status,
                str(len(pods)),
                "\n".join(pod_lines),
                "\n".join(volumes_display) or "-",
                timestamp,
                ttl,
                creator,
                queue,
                cpu or "—",
                memory or "—",
                gpu or "—",
            )

        table.pad_bottom = 1
        console.print(table)

    except ApiException as e:
        console.print(f"[red]Kubernetes API error: {e}[/red]")
        raise typer.Exit(1)


@app.command("port-forward")
def kt_port_forward(
    name: str = service_name_argument(help="Service name"),
    local_port: int = typer.Argument(default=serving_constants.DEFAULT_KT_SERVER_PORT, help="Local port to bind to"),
    remote_port: int = typer.Argument(
        default=serving_constants.DEFAULT_KT_SERVER_PORT,
        help="Remote port to forward to",
    ),
    namespace: str = typer.Option(
        globals.config.namespace,
        "-n",
        "--namespace",
    ),
    pod: str = typer.Option(
        None,
        "-p",
        "--pod",
        help="Name or index of a specific pod to load logs from (0-based)",
    ),
):
    """
    Port forward a local port to the specified Kubetorch service.

    Examples:

    .. code-block:: bash


        $ kt port-forward my-service

        $ kt port-forward my-service 32300

        $ kt port-forward my-service -n custom-namespace

        $ kt port-forward my-service -p my-pod

    This allows you to access the service locally using `curl http://localhost:<port>`.
    """

    from kubetorch.resources.compute.utils import is_port_available

    if not is_port_available(local_port):
        console.print(f"\n[red]Local port {local_port} is already in use.[/red]")
        raise typer.Exit(1)

    core_api, custom_api, apps_v1_api = initialize_k8s_clients()

    name, _ = get_deployment_mode(name, namespace, custom_api, apps_v1_api)
    pods = validate_pods_exist(name, namespace, core_api)
    sorted_by_time = sorted(pods, key=lambda pod: pod.metadata.creation_timestamp)

    if pod:  # case when the user provides a pod
        pod_name = validate_provided_pod(service_name=name, provided_pod=pod, service_pods=sorted_by_time)
    else:  # if user does not provide pod, port-forward to the first pod by default
        pod_name = sorted_by_time[0].metadata.name

    process = None

    def cleanup_process():
        # Clean up the port forward process
        if process:
            process.kill()

    def signal_handler(signum, frame):
        """Handle interrupt signals for graceful shutdown."""
        console.print(f"\nReceived signal {signum}, cleaning up port forward...")
        cleanup_process()
        console.print("Port forward stopped.")
        raise typer.Exit(0)

    # Register signal handlers for graceful shutdown
    signal.signal(signal.SIGINT, signal_handler)
    signal.signal(signal.SIGTERM, signal_handler)

    from kubetorch.serving.utils import wait_for_port_forward

    cmd = [
        "kubectl",
        "port-forward",
        f"pod/{pod_name}",
        f"{local_port}:{remote_port}",
        "--namespace",
        namespace,
    ]

    port_forward_msg = f"Starting port forward to {name} in namespace {namespace}"

    if pod:
        port_forward_msg = port_forward_msg + f", pod: [reset]{pod}"
    console.print(port_forward_msg)

    process = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE, start_new_session=True)

    try:
        wait_for_port_forward(process, local_port)
        time.sleep(2)
    except Exception as e:
        logger.info(f"Failed to establish port forward on port {local_port}: {e}")
        if process:
            cleanup_process()
            process = None
        return

    console.print(f"[green]✓ Port forward active on localhost:{local_port} -> {pod_name}:{remote_port}[/green]")
    console.print(f"[cyan]You can now run: curl http://localhost:{local_port}[/cyan]")
    console.print("[dim]Press Ctrl+C to stop the port forward[/dim]")

    # Keep the port forward running until interrupted
    try:
        while True:
            if process.poll() is not None:
                # Process has terminated
                console.print("[red]Port forward process has terminated unexpectedly[/red]")
                break
            time.sleep(1)
    except KeyboardInterrupt:
        # This should be handled by the signal handler, but just in case
        pass

    except typer.Exit:
        # Re-raise typer.Exit to maintain proper CLI behavior
        raise
    except Exception as e:
        console.print(f"[red]Error during port forwarding: {e}[/red]")
        raise typer.Exit(1)
    finally:
        cleanup_process()


@app.command("run", context_settings={"allow_extra_args": True, "ignore_unknown_options": True})
def kt_run(
    ctx: typer.Context,
    name: str = typer.Option(None, "--name", help="Name for the run"),
    run_async: bool = typer.Option(False, "--async", help="Whether to run async and not stream logs live"),
    file: int = typer.Option(None, "--file", help="File where the app is defined in"),
):
    """
    Build and deploy a kubetorch app that runs the provided CLI command. In order for the app
    to be deployed, the file being run must be a Python file specifying a `kt.app` construction
    at the top of the file.

    Examples:

    .. code-block:: bash

        $ kt run python train.py --epochs 5
        $ kt run fastapi run my_app.py --name fastapi-app
    """
    from kubetorch import App

    cli_cmd = " ".join(ctx.args)
    if not cli_cmd:
        raise typer.BadParameter("You must provide a command to run.")
    elif cli_cmd.split()[0].endswith(".py"):
        raise typer.BadParameter(
            "You must provide a full command to run, the Python file should not be the first argument. "
            "(e.g. `kt run python train.py`)"
        )

    python_file = file
    if not python_file:
        for arg in cli_cmd.split():
            if arg.endswith("py") and Path(arg).exists():
                python_file = arg
                break

        if not python_file:
            console.print(
                f"[red]Could not detect python file with `kt.app` in {cli_cmd}. Pass it in with `--file`.[/red]"
            )
            raise typer.Exit(1)

    # Set env vars for construction of app instance
    os.environ["KT_RUN"] = "1"
    os.environ["KT_RUN_CMD"] = cli_cmd
    os.environ["KT_RUN_FILE"] = python_file
    if name:
        os.environ["KT_RUN_NAME"] = name
    if run_async:
        os.environ["KT_RUN_ASYNC"] = "1"

    # Extract the app instance from the python file
    module_name = Path(python_file).stem
    python_file_dir = Path(python_file).resolve().parent

    # Add the directory containing the Python file to sys.path to support relative imports
    if str(python_file_dir) not in sys.path:
        sys.path.insert(0, str(python_file_dir))

    spec = importlib.util.spec_from_file_location(module_name, python_file)
    module = importlib.util.module_from_spec(spec)
    sys.modules[module_name] = module
    spec.loader.exec_module(module)

    app_instance = None
    for _, obj in inspect.getmembers(module):
        if isinstance(obj, App):
            app_instance = obj
            break
    if not app_instance:
        console.print(f"[red]Could not find kt.app definition in {python_file} [/red]")
        raise typer.Exit(1)

    app_instance.deploy()


@app.command("secrets")
def kt_secrets(
    action: SecretAction = typer.Argument(
        SecretAction.list,
        help="Action to perform: list, create, update, delete, describe",
    ),
    name: str = typer.Argument(None, help="Secret name (for create or delete actions)"),
    prefix: str = typer.Option(
        None,
        "--prefix",
        "-x",
    ),
    namespace: str = typer.Option(
        "default",
        "-n",
        "--namespace",
    ),
    all_namespaces: bool = typer.Option(
        False,
        "--all-namespaces",
        "-A",
    ),
    yes: bool = typer.Option(False, "-y", "--yes", help="Deletion confirmation"),
    path: str = typer.Option(None, "--path", "-p", help="Path where the secret values are held"),
    provider: str = typer.Option(
        None,
        "--provider",
        "-c",
        help="Provider corresponding to the secret (e.g. 'aws', 'gcp'). "
        "If not specified, secrets are loaded from the default provider path.",
    ),
    env_vars: List[str] = typer.Option(
        None,
        "--env-vars",
        "-v",
        help="Environment variable(s) key(s) whose value(s) will hold the secret value(s)",
    ),
    show_values: bool = typer.Option(False, "-s", "--show", help="Show secrets values in the describe output"),
):
    """Manage secrets used in Kubetorch services.

    Examples:

    .. code-block:: bash

        $ kt secrets  # list secrets in the default namespace

        $ kt secrets list -n my_namespace  # list secrets in `my_namespace` namespace

        $ kt secrets -A  # list secrets in all namespaces

        $ kt secrets create --provider aws  # create a secret with the aws credentials in `default` namespace

        $ kt secrets create my_secret -v ENV_VAR_1 -v ENV_VAR_2 -n my_namespace  # create a secret using env vars

        $ kt secrets delete my_secret -n my_namespace  # delete a secret called `my_secret` from `my_namespace` namespace

        $ kt secrets delete aws   # delete a secret called `aws` from `default` namespace
    """
    import kubetorch as kt
    from kubetorch.resources.compute.utils import delete_secrets, list_secrets
    from kubetorch.resources.secrets.kubernetes_secrets_client import KubernetesSecretsClient

    secrets_client = KubernetesSecretsClient(namespace=namespace)

    core_api, custom_api, apps_v1_api = initialize_k8s_clients()

    if action == SecretAction.list:
        secrets = list_secrets(
            core_api=core_api,
            namespace=namespace,
            prefix=prefix,
            all_namespaces=all_namespaces,
            console=console,
            filter_by_creator=False,
        )

        table_columns = [
            ("SECRET", "blue"),
            ("CREATOR", "cyan"),
            ("NAMESPACE", "yellow"),
        ]
        table = create_table_for_output(
            columns=table_columns,
            no_wrap_columns_names=["SECRET"],
            header_style={"bold": True},
        )

        if not secrets:
            msg = "No secrets found"
            if not all_namespaces:
                if prefix:
                    msg += f" with prefix: {prefix}"
                msg += f" in namespace: {namespace}"
            console.print(f"[yellow]{msg}[/yellow]")
            raise typer.Exit(0)

        for secret in secrets:
            secret_name = secret.get(
                "user_defined_name"
            )  # TODO: maybe display the kt name? so it'll match kubectl get secrets
            creator = secret.get("username")
            namespace = secret.get("namespace")
            table.add_row(secret_name, creator, namespace)

        table.pad_bottom = 1
        console.print(table)

    elif action == SecretAction.create:
        if not (name or provider):
            console.print("[red]Cannot create secret: name or provider must be specified.[/red]")
            typer.Exit(1)
        env_vars_dict = {key: key for key in env_vars} if env_vars else {}

        try:
            new_secret = kt.secret(name=name, provider=provider, path=path, env_vars=env_vars_dict)
            secrets_client.create_secret(secret=new_secret, console=console)
        except Exception as e:
            console.print(f"[red]Failed to create the secret: {e}[/red]")
            raise typer.Exit(0)

    elif action == SecretAction.delete:
        prefix = name if name else prefix
        all_namespaces = False if name else all_namespaces
        secrets_to_delete = list_secrets(
            core_api=core_api,
            namespace=namespace,
            prefix=prefix,
            all_namespaces=all_namespaces,
            console=console,
        )

        username = globals.config.username
        secrets_to_delete_by_namespace: dict[str, list[str]] = {}
        for secret in secrets_to_delete:
            ns = secret.get("namespace")
            name = secret.get("name")

            if all_namespaces:
                if secret.get("username") != username:
                    continue  # skip secrets not owned by user

            secrets_to_delete_by_namespace.setdefault(ns, []).append(name)

        # Flatten names for display
        secrets_names = [name for names in secrets_to_delete_by_namespace.values() for name in names]

        if not secrets_names:
            console.print(f"[yellow]No secrets to delete for username: {username}[/yellow]")
            raise typer.Exit(0)

        secrets_word = "secret" if len(secrets_names) == 1 else "secrets"
        console.print(f"\nDeleting {len(secrets_names)} {secrets_word}...")

        for secret in secrets_names:
            console.print(f"  - [blue]{secret}[/blue]")

        if not yes:
            confirm = typer.confirm("\nDo you want to proceed?")
            if not confirm:
                console.print("[yellow]Operation cancelled[/yellow]")
                raise typer.Exit(0)

        for ns, secrets in secrets_to_delete_by_namespace.items():
            if secrets:
                client = KubernetesSecretsClient(namespace=ns)
                delete_secrets(
                    secrets=secrets,
                    console=console,
                    secrets_client=client,
                )

    elif action == SecretAction.describe:
        prefix = name if name else prefix
        all_namespaces = False if name else all_namespaces
        secrets_to_describe = list_secrets(
            core_api=core_api,
            namespace=namespace,
            prefix=name or prefix,
            all_namespaces=all_namespaces,
            filter_by_creator=False,
            console=console,
        )
        if not secrets_to_describe:
            console.print("[yellow] No secrets found[/yellow]")
            raise typer.Exit(0)

        for secret in secrets_to_describe:
            k8_name = secret.get("name")
            kt_name = secret.get("user_defined_name")
            console.print(f"[bold cyan]{kt_name}[/bold cyan]")
            console.print(f"  K8 Name: [reset]{k8_name}")
            console.print(f'  Namespace: {secret.get("namespace")}')
            console.print(f'  Labels: [reset]{secret.get("labels")}')
            console.print(f'  Type: {secret.get("type")}')
            secret_data = secret.get("data")
            if show_values:
                console.print("  Data:")
                for k, v in secret_data.items():
                    try:
                        decoded_value = base64.b64decode(v).decode("utf-8")
                    except Exception:
                        decoded_value = "<binary data>"
                    indented_value = textwrap.indent(decoded_value, "     ")
                    indented_value = indented_value.replace("\n\n", "\n")
                    console.print(f"   {k}:{indented_value}\n")


@app.command("ssh")
def kt_ssh(
    name: str = service_name_argument(help="Service name"),
    namespace: str = typer.Option(
        globals.config.namespace,
        "-n",
        "--namespace",
    ),
    pod: str = typer.Option(
        None,
        "-p",
        "--pod",
        help="Name or index of a specific pod to load logs from (0-based)",
    ),
):
    """SSH into a remote service. By default, will SSH into the first pod.

    Examples:

    .. code-block:: bash

        $ kt ssh my_service
    """
    core_api, custom_api, apps_v1_api = initialize_k8s_clients()

    try:
        # Validate service exists and get deployment mode
        name, deployment_mode = get_deployment_mode(name, namespace, custom_api, apps_v1_api)

        # Get and validate pods
        pods = validate_pods_exist(name, namespace, core_api)

        sorted_by_time = sorted(pods, key=lambda pod: pod.metadata.creation_timestamp)

        # case when the user provides a specific pod to ssh into
        if pod:
            pod_name = validate_provided_pod(service_name=name, provided_pod=pod, service_pods=sorted_by_time)
        # if pod is not provided, ssh into the first pod.
        else:
            pod_name = sorted_by_time[0].metadata.name

        console.print(f"[green]Found pod:[/green] [blue]{pod_name}[/blue] ({deployment_mode})")
        console.print("[yellow]Connecting to pod...[/yellow]")

        # Still need subprocess for the interactive shell
        subprocess.run(
            ["kubectl", "exec", "-it", pod_name, "-n", namespace, "--", "/bin/bash"],
            check=True,
        )

    except ApiException as e:
        console.print(f"[red]Kubernetes API error: {e}[/red]")
        raise typer.Exit(1)


@app.command("teardown")
def kt_teardown(
    name: str = service_name_argument(help="Service name", required=False),
    yes: bool = typer.Option(False, "-y", "--yes", help="Deletion confirmation"),
    teardown_all: bool = typer.Option(False, "-a", "--all", help="Deletes all services for the current user"),
    prefix: str = typer.Option("", "-p", "--prefix", help="Tear down all services with given prefix"),
    namespace: str = typer.Option(
        globals.config.namespace,
        "-n",
        "--namespace",
    ),
    force: bool = typer.Option(False, "-f", "--force", help="Force deletion without graceful shutdown"),
    exact_match: bool = typer.Option(
        False,
        "-e",
        "--exact-match",
        help="Only delete the exact service name, not the prefixed version",
    ),
):
    """Delete a service and all its associated resources (deployments, configmaps, etc).


    Examples:

    .. code-block:: bash

        $ kt teardown my-service -y  # force teardown resources corresponding to service

        $ kt teardown --all          # teardown all resources corresponding to username

        $ kt teardown --prefix test  # teardown resources with prefix "test"
    """
    from kubetorch import config
    from kubetorch.resources.compute.utils import delete_resources_for_service, fetch_resources_for_teardown

    name, yes, teardown_all, namespace, prefix = default_typer_values(name, yes, teardown_all, namespace, prefix)

    core_api, custom_api, _ = initialize_k8s_clients()

    if teardown_all:
        if not config.username:
            console.print(
                "[red]Username is not found, can't delete all services. Please set up a username, provide a service "
                "name or use the --prefix flag[/red]"
            )
            raise typer.Exit(1)

        console.print(f"Deleting all services for username [blue]{config.username}[/blue]...")

    elif prefix:
        console.print(
            f"Deleting all services with prefix [blue]{prefix}[/blue] in [blue]{namespace}[/blue] namespace..."
        )
    else:
        if not name:
            console.print("[red]Please provide a service name or use the --all or --prefix flags[/red]")
            raise typer.Exit(1)

        console.print(f"Finding resources for service [blue]{name}[/blue] in [blue]{namespace}[/blue] namespace...")

    resources = fetch_resources_for_teardown(
        namespace=namespace,
        target=name,
        core_api=core_api,
        custom_api=custom_api,
        prefix=prefix,
        username=config.username if teardown_all else None,
        exact_match=exact_match,
    )

    services = list(resources["services"].keys())
    service_count = len(services)

    if teardown_all or prefix:
        service_word = "service" if service_count == 1 else "services"
        if not services:
            console.print("[yellow]No services are found[/yellow]")
            raise typer.Exit(0)
        else:
            console.print(f"[yellow]Found [bold]{service_count}[/bold] {service_word} to delete.[/yellow]")

    if name and not services:
        console.print(f"[red]Service [bold]{name}[/bold] not found[/red]")
        raise typer.Exit(1)

    # Confirmation prompt for multiple services
    if not yes and service_count > 1:
        for service_name in services:
            console.print(f" • {service_name}")

    # List out resources to be deleted for each service
    console.print("\n[yellow]The following resources will be deleted:[/yellow]")
    for name in services:
        service_info = resources["services"][name]
        configmaps = service_info["configmaps"]
        service_type = service_info.get("type", "unknown")

        if service_type == "deployment":
            console.print(f"• Deployment: [blue]{name}[/blue]")
            console.print(f"• Service: [blue]{name}[/blue]")
        elif service_type == "knative":
            console.print(f"• Knative Service: [blue]{name}[/blue]")
        elif service_type == "raycluster":
            console.print(f"• RayCluster: [blue]{name}[/blue]")
            console.print(f"• Service: [blue]{name}[/blue]")
        else:
            console.print(f"• Service: [blue]{name}[/blue]")

        if configmaps:
            console.print("• ConfigMaps:")
            for cm in configmaps:
                console.print(f"  - [blue]{cm}[/blue]")

    # Confirmation prompt for single service
    if not yes and not force:  # if --force is provided, we don't need additional confirmation
        confirm = typer.confirm("\nDo you want to proceed?")
        if not confirm:
            console.print("[yellow]Teardown cancelled[/yellow]")
            raise typer.Exit(0)

    # Delete resources
    if force:
        console.print("\n[yellow]Force deleting resources...[/yellow]")
    else:
        console.print("\n[yellow]Deleting resources...[/yellow]")

    service_types = set()
    for name in services:
        service_info = resources["services"][name]
        configmaps = service_info["configmaps"]
        service_type = service_info.get("type", "knative")
        service_types.add(service_type)

        delete_resources_for_service(
            core_api=core_api,
            custom_api=custom_api,
            configmaps=configmaps,
            name=name,
            service_type=service_type,
            namespace=namespace,
            console=console,
            force=force,
        )

    # Force delete any remaining pods if --force flag is set
    if force:
        # Build list of service names to check for pods
        # Include both found services and the original target name (in case service was already deleted)
        service_names_to_check = list(services)
        if name and name not in service_names_to_check:
            service_names_to_check.append(name)

        if service_names_to_check:
            console.print("\n[yellow]Force deleting any remaining pods...[/yellow]")
            for service_name in service_names_to_check:
                try:
                    # Get pods matching the service
                    pods = core_api.list_namespaced_pod(
                        namespace=namespace,
                        label_selector=f"kubetorch.com/service={service_name}",
                    ).items

                    if pods:
                        for pod in pods:
                            try:
                                core_api.delete_namespaced_pod(
                                    name=pod.metadata.name,
                                    namespace=namespace,
                                    grace_period_seconds=0,
                                    propagation_policy="Background",
                                )
                                console.print(f"✓ Force deleted pod [blue]{pod.metadata.name}[/blue]")
                            except ApiException as e:
                                if e.status != 404:  # Ignore if already deleted
                                    console.print(f"[red]Failed to delete pod {pod.metadata.name}: {e}[/red]")
                except Exception as e:
                    console.print(f"[red]Failed to list pods for service {service_name}: {e}[/red]")

        # Also check for any orphaned pods with kubetorch labels if using --all or --prefix
        if teardown_all or prefix:
            try:
                label_selector = "kubetorch.com/service"
                if teardown_all and config.username:
                    label_selector += f",kubetorch.com/username={config.username}"

                all_pods = core_api.list_namespaced_pod(namespace=namespace, label_selector=label_selector).items

                # Filter by prefix if specified
                if prefix:
                    all_pods = [
                        p for p in all_pods if p.metadata.labels.get("kubetorch.com/service", "").startswith(prefix)
                    ]

                # Delete any remaining pods not already handled
                for pod in all_pods:
                    if pod.metadata.name not in [
                        p.metadata.name
                        for s in service_names_to_check
                        for p in core_api.list_namespaced_pod(
                            namespace=namespace,
                            label_selector=f"kubetorch.com/service={s}",
                        ).items
                    ]:
                        try:
                            core_api.delete_namespaced_pod(
                                name=pod.metadata.name,
                                namespace=namespace,
                                grace_period_seconds=0,
                                propagation_policy="Background",
                            )
                            console.print(f"✓ Force deleted orphaned pod [blue]{pod.metadata.name}[/blue]")
                        except ApiException as e:
                            if e.status != 404:
                                console.print(f"[red]Failed to delete orphaned pod {pod.metadata.name}: {e}[/red]")
            except Exception as e:
                console.print(f"[red]Failed to list orphaned pods: {e}[/red]")

    console.print("\n[green]Teardown completed successfully[/green]")


@app.command("volumes")
def kt_volumes(
    action: VolumeAction = typer.Argument(VolumeAction.list, help="Action to perform"),
    name: str = typer.Argument(None, help="Volume name (for create action)"),
    storage_class: str = typer.Option(None, "--storage-class", "-c", help="Storage class"),
    size: str = typer.Option("10Gi", "--size", "-s", help="Volume size (default: 10Gi)"),
    access_mode: str = typer.Option("ReadWriteMany", "--access-mode", "-a", help="Access mode"),
    mount_path: str = typer.Option(
        None,
        "--mount-path",
        "-m",
        help=f"Mount path (default: /{KT_MOUNT_FOLDER}/{{name}})",
    ),
    namespace: str = typer.Option(
        globals.config.namespace,
        "-n",
        "--namespace",
    ),
    all_namespaces: bool = typer.Option(
        False,
        "--all-namespaces",
        "-A",
        help="List volumes across all namespaces",
    ),
):
    """Manage volumes used in Kubetorch services.

    Examples:

    .. code-block:: bash

        $ kt volumes

        $ kt volumes -A

        $ kt volumes create my-vol

        $ kt volumes create my-vol -c gp3-csi -s 20Gi

        $ kt volumes delete my-vol

        $ kt volumes ssh my-vol
    """
    from kubernetes import client

    from kubetorch import Volume
    from kubetorch.utils import load_kubeconfig

    load_kubeconfig()
    core_v1 = client.CoreV1Api()

    target_namespace = None
    if not all_namespaces:
        target_namespace = namespace or globals.config.namespace

    if action == VolumeAction.list:
        try:
            if all_namespaces:
                pvcs = core_v1.list_persistent_volume_claim_for_all_namespaces()
                title = "Kubetorch Volumes (All Namespaces)"
            else:
                pvcs = core_v1.list_namespaced_persistent_volume_claim(namespace=target_namespace)
                title = f"Kubetorch Volumes (Namespace: {target_namespace})"

            # List all Kubetorch PVCs
            kubetorch_pvcs = [
                pvc for pvc in pvcs.items if (pvc.metadata.annotations or {}).get("kubetorch.com/mount-path")
            ]

            if not kubetorch_pvcs:
                if all_namespaces:
                    console.print("[yellow]No volumes found in all namespaces[/yellow]")
                else:
                    console.print(f"[yellow]No volumes found in namespace {target_namespace}[/yellow]")
                return

            table = Table(title=title)
            if all_namespaces:
                table.add_column("Namespace", style="green")
            table.add_column("Name", style="cyan")
            table.add_column("PVC Name", style="blue")
            table.add_column("Status", style="green")
            table.add_column("Size", style="yellow")
            table.add_column("Storage Class", style="magenta")
            table.add_column("Access Mode", style="white")
            table.add_column("Mount Path", style="dim")

            for pvc in kubetorch_pvcs:
                # Extract volume name from PVC name
                volume_name = pvc.metadata.name
                status = pvc.status.phase
                size = pvc.spec.resources.requests.get("storage", "Unknown")
                storage_class = pvc.spec.storage_class_name or "Default"
                access_mode = pvc.spec.access_modes[0] if pvc.spec.access_modes else "Unknown"

                # Get mount path from annotations
                annotations = pvc.metadata.annotations or {}
                mount_path_display = annotations.get("kubetorch.com/mount-path", f"/{KT_MOUNT_FOLDER}/{volume_name}")

                status_color = "green" if status == "Bound" else "yellow" if status == "Pending" else "red"

                row_data = []
                if all_namespaces:
                    row_data.append(pvc.metadata.namespace)

                row_data.extend(
                    [
                        volume_name,
                        pvc.metadata.name,
                        f"[{status_color}]{status}[/{status_color}]",
                        size,
                        storage_class,
                        access_mode,
                        mount_path_display,
                    ]
                )

                table.add_row(*row_data)

            console.print(table)

        except Exception as e:
            console.print(f"[red]Failed to list volumes: {e}[/red]")
            raise typer.Exit(1)

    elif action == VolumeAction.ssh:
        if not name:
            console.print("[red]Volume name is required[/red]")
            raise typer.Exit(1)

        volume = Volume.from_name(name=name, namespace=namespace, core_v1=core_v1)
        volume.ssh()

    elif action == VolumeAction.create:
        if not name:
            console.print("[red]Volume name is required[/red]")
            raise typer.Exit(1)

        if all_namespaces:
            console.print("[red]Cannot create volume with --all-namespaces. Specify a namespace.[/red]")
            raise typer.Exit(1)

        try:
            volume = Volume(
                name=name,
                storage_class=storage_class,
                mount_path=mount_path,
                size=size,
                access_mode=access_mode,
                namespace=namespace,
            )

            if volume.exists():
                console.print(
                    f"[yellow]Volume {name} (PVC: {volume.pvc_name}) already exists in "
                    f"namespace {namespace}[/yellow]"
                )
                return

            console.print(f"Creating volume [blue]{name}[/blue]...")
            volume.create()

            console.print(f"[green]✓[/green] Successfully created volume [blue]{name}[/blue]")
            config = volume.config()
            for k, v in config.items():
                console.print(f"[bold]• {k}[/bold]: {v}")

        except Exception as e:
            console.print(f"[red]Failed to create volume {name}: {e}[/red]")
            raise typer.Exit(1)

    elif action == VolumeAction.delete:
        if not name:
            console.print("[red]Volume name is required[/red]")
            raise typer.Exit(1)

        if all_namespaces:
            console.print("[red]Cannot delete volume with --all-namespaces. Specify a namespace.[/red]")
            raise typer.Exit(1)

        try:
            volume = Volume.from_name(name=name, namespace=namespace, core_v1=core_v1)

            console.print(f"Deleting volume [blue]{name}[/blue]...")
            volume.delete()

            console.print(f"[green]✓[/green] Successfully deleted volume [blue]{name}[/blue]")

        except ValueError:
            console.print(f"[red]Volume {name} not found in namespace {namespace}[/red]")
            raise typer.Exit(1)

        except Exception as e:
            console.print(f"[red]Failed to delete volume {name}: {e}[/red]")
            raise typer.Exit(1)


@app.callback(invoke_without_command=True, help="Kubetorch CLI")
def main(
    ctx: typer.Context,
    version: bool = typer.Option(None, "--version", "-v", help="Show the version and exit."),
):
    if version:
        from kubetorch import __version__

        print(f"{__version__}")
    elif ctx.invoked_subcommand is None:
        subprocess.run("kubetorch --help", shell=True)



import os
import time
from pathlib import Path
from typing import List, Optional

import yaml
from kubernetes import client, config
from kubernetes.client import V1Pod
from kubernetes.stream import stream

from kubetorch.constants import DEFAULT_KUBECONFIG_PATH
from kubetorch.globals import config as kt_config

from kubetorch.logger import get_logger
from kubetorch.servers.http.utils import is_running_in_kubernetes

logger = get_logger(__name__)


def get_k8s_identity_name() -> Optional[str]:
    """Get Kubernetes user identity from kubeconfig file.

    Returns:
        User identity string (e.g., "user-{name}", "role-{name}", "sa-{name}") or None
    """
    try:
        if is_running_in_kubernetes():
            # Get the service account name
            service_account_name = os.environ.get("SERVICE_ACCOUNT_NAME")
            if service_account_name:
                return "sa-" + service_account_name.lower()
            return None

        # Read from kubeconfig
        kubeconfig_path = os.getenv("KUBECONFIG") or DEFAULT_KUBECONFIG_PATH
        kubeconfig_file = Path(kubeconfig_path).expanduser()
        if not kubeconfig_file.exists():
            return None

        with open(kubeconfig_file, "r") as f:
            kubeconfig = yaml.safe_load(f)

        current_context = kubeconfig.get("current-context")
        if not current_context:
            return None

        # Find current context's user
        for context in kubeconfig.get("contexts", []):
            if context.get("name") == current_context:
                user_name = context.get("context", {}).get("user")
                if not user_name:
                    return None

                # Parse AWS ARN format (EKS IAM users/roles)
                if "assumed-role" in user_name:
                    parts = user_name.split("/")
                    if len(parts) >= 2:
                        return "role-" + parts[-2].lower()
                elif "/" in user_name and (".amazonaws.com" in user_name or "arn:aws" in user_name):
                    parts = user_name.split("/")
                    return "user-" + parts[-1].lower()

                # Check for exec-based auth with AWS role
                for user in kubeconfig.get("users", []):
                    if user.get("name") == user_name:
                        exec_config = user.get("user", {}).get("exec", {})
                        for env_var in exec_config.get("env", []):
                            if env_var.get("name") == "AWS_ROLE_ARN":
                                role_arn = env_var.get("value", "")
                                if "/" in role_arn:
                                    return "role-" + role_arn.split("/")[-1].lower()
                        break

                # Default: use user name as-is
                return "user-" + user_name.lower()

    except Exception as e:
        logger.debug(f"Failed to get Kubernetes identity name: {e}")

    return None


def read_files_as_secrets_dict(path: str, filenames: List[str]):
    values = {}
    cred_path = os.path.expanduser(path)

    for filename in filenames:
        file_path = os.path.join(cred_path, filename)
        # Read the files
        content = _read_file_if_exists(file_path)
        if content:
            values[filename] = content
        # # Base64 encode the content
        # encoded = base64.b64encode(content).decode("utf-8")
        # values[filename] = encoded

    return values


def _read_file_if_exists(file_path: str) -> Optional[str]:
    try:
        with open(file_path, "r") as f:  # "rb" if you encode above.
            return f.read()
    except FileNotFoundError:
        logger.error(f"Warning: {file_path} not found, using empty content")
        return None


# ------------------------------------------------------------------------------------------------
# Secret testing utils
# ------------------------------------------------------------------------------------------------


def check_path_on_kubernetes_pods(path: str, service_name: str, namespace: str = None) -> bool:
    """
    Check if a path exists on a specific Knative service's pods
    """
    namespace = namespace or kt_config.namespace
    # Load Kubernetes configuration
    config.load_kube_config()
    # Initialize API clients
    core_v1_api = client.CoreV1Api()

    pods = _fetch_pods_for_kubernetes_service(service_name, namespace, core_v1_api)
    if not pods:
        logger.error(f"No pods found for service {service_name} in namespace {namespace}")
        return False

    path_found = True
    for pod in pods:
        pod_name = pod.metadata.name
        command = ["/bin/bash", "-c", f"[ -f {path} ] && echo yes || echo no"]
        try:
            resp = stream(
                core_v1_api.connect_get_namespaced_pod_exec,
                name=pod_name,
                namespace=namespace,
                command=command,
                container="kubetorch",
                stderr=True,
                stdout=True,
            )
            if "yes" in resp:
                continue
        except client.exceptions.ApiException as e:
            logger.error(f"Error executing command on pod {pod_name}: {e}")

        path_found = False

    return path_found


def check_env_vars_on_kubernetes_pods(env_vars: list, service_name: str, namespace: str = None) -> dict:
    """
    Check if an AWS role is assumed on a specific Knative service's pods

    :param namespace: Kubernetes namespace
    :param service_name: Name of the Knative service
    :return: Dictionary with role assumption details
    """
    namespace = namespace or kt_config.namespace
    # Load Kubernetes configuration
    config.load_kube_config()
    # Initialize API clients
    core_v1_api = client.CoreV1Api()

    pods = _fetch_pods_for_kubernetes_service(service_name, namespace, core_v1_api)
    if not pods:
        logger.error(f"No pods found for service {service_name} in namespace {namespace}")
        return {}

    found_env_vars = {}

    for pod in pods:
        for env_var in env_vars:
            if found_env_vars.get(env_var):
                # Skip if already found on another pod
                continue
            pod_name = pod.metadata.name
            command = ["/bin/bash", "-c", f"echo ${env_var}"]
            try:
                resp = stream(
                    core_v1_api.connect_get_namespaced_pod_exec,
                    name=pod_name,
                    namespace=namespace,
                    command=command,
                    container="kubetorch",
                    stderr=True,
                    stdout=True,
                )
                if len(resp.strip()) > 0:
                    found_env_vars[env_var] = resp.strip()
            except client.exceptions.ApiException as e:
                logger.error(f"Error executing command: {e}")

        if set(found_env_vars.keys()) == set(env_vars):
            # Found all env vars: skip the remaining pods
            break

    return found_env_vars


def _fetch_pods_for_kubernetes_service(service_name: str, namespace: str, client_api: client.CoreV1Api) -> List[V1Pod]:
    """
    Fetch pods for a specific Knative service with timeout
    """
    start_time = time.time()
    while time.time() - start_time < 30:
        try:
            # List pods matching the service
            pods = client_api.list_namespaced_pod(
                namespace=namespace,
                label_selector=f"kubetorch.com/service={service_name}",
            )
            ready_pods = [pod for pod in pods.items if pod.status.phase == "Running"]
            if ready_pods:
                return ready_pods
        except Exception as e:
            logger.error(f"Error fetching pods for service {service_name} in namespace {namespace}: {e}")
        time.sleep(1)

    return []



from .. import Secret


class AWSSecret(Secret):
    """
    .. note::
            To create an AWSSecret, please use the factory method :func:`secret` with ``provider="aws"``.
    """

    _PROVIDER = "aws"
    _DEFAULT_PATH = "~/.aws"
    _DEFAULT_FILENAMES = ["config", "credentials"]
    _DEFAULT_ENV_VARS = {
        "access_key": "AWS_ACCESS_KEY_ID",
        "secret_key": "AWS_SECRET_ACCESS_KEY",
    }



LOCALHOST: str = "127.0.0.1"
DEFAULT_KUBECONFIG_PATH = "~/.kube/config"
MAX_PORT_TRIES = 10

# CLI constants
DOUBLE_SPACE_UNICODE = "\u00A0\u00A0"
BULLET_UNICODE = "\u2022"

MAX_USERNAME_LENGTH = 16

CPU_RATE = 0.01
GPU_RATE = 0.05

KT_MOUNT_FOLDER = "ktfs"
DEFAULT_VOLUME_ACCESS_MODE = "ReadWriteMany"

DASHBOARD_PORT = 3001
GRAFANA_PORT = 3000



import asyncio
import json
import tempfile
import threading
import time
import urllib.parse
from datetime import datetime, timezone
from pathlib import Path
from typing import Dict, List, Union

import websockets

from kubetorch.globals import config, service_url, service_url_async
from kubetorch.logger import get_logger
from kubetorch.resources.callables.utils import get_names_for_reload_fallbacks, locate_working_dir

from kubetorch.resources.compute.utils import (
    delete_cached_service_data,
    delete_configmaps,
    load_configmaps,
    VersionMismatchError,
)
from kubetorch.servers.http.http_client import HTTPClient
from kubetorch.servers.http.utils import (
    clean_and_validate_k8s_name,
    generate_unique_request_id,
    is_running_in_kubernetes,
)
from kubetorch.serving.utils import has_k8s_credentials, KubernetesCredentialsError
from kubetorch.utils import (
    extract_host_port,
    get_kt_install_url,
    iso_timestamp_to_nanoseconds,
    LogVerbosity,
    ServerLogsFormatter,
)

logger = get_logger(__name__)


class Module:
    MODULE_TYPE = None

    def __init__(
        self,
        name: str,
        pointers: tuple,
    ):
        self._compute = None
        self._deployment_timestamp = None
        self._service_config = None
        self._http_client = None
        self._get_if_exists = True
        self._reload_prefixes = None
        self._serialization = "json"  # Default serialization format
        self._async = False
        self._remote_pointers = None
        self._service_name = None

        self.pointers = pointers
        self.name = clean_and_validate_k8s_name(name, allow_full_length=False) if name else None

    @property
    def module_name(self):
        """Name of the function or class."""
        return self.pointers[2]

    @property
    def reload_prefixes(self):
        return self._reload_prefixes or []

    @reload_prefixes.setter
    def reload_prefixes(self, value: Union[str, List[str]]):
        """Set the reload_prefixes property."""
        if isinstance(value, (list)):
            self._reload_prefixes = value
        elif isinstance(value, str):
            self._reload_prefixes = [value]
        else:
            raise ValueError("`reload_prefixes` must be a string or a list.")

    @property
    def namespace(self):
        """Namespace where the service is deployed."""
        if self.compute is not None:
            return self.compute.namespace
        return config.namespace

    @property
    def service_name(self):
        """Name of the knative service, formatted according to k8s regex rules."""
        if self._service_name:
            return self._service_name

        service_name = self.name

        if config.username and not self.reload_prefixes and not service_name.startswith(config.username + "-"):
            service_name = f"{config.username}-{service_name}"

        self._service_name = clean_and_validate_k8s_name(service_name, allow_full_length=True)
        return self._service_name

    @service_name.setter
    def service_name(self, value: str):
        self._service_name = clean_and_validate_k8s_name(value, allow_full_length=True)

    @property
    def compute(self):
        """Compute object corresponding to the module."""
        return self._compute

    @compute.setter
    def compute(self, compute: "Compute"):
        self._compute = compute

    @property
    def deployment_timestamp(self):
        if not self._deployment_timestamp:
            self._deployment_timestamp = self.compute.service_manager.get_deployment_timestamp_annotation(
                self.service_name
            )
        return self._deployment_timestamp

    @deployment_timestamp.setter
    def deployment_timestamp(self, value: str):
        self._deployment_timestamp = value

    @property
    def remote_pointers(self):
        if self._remote_pointers:
            return self._remote_pointers

        source_dir, _ = locate_working_dir(self.pointers[0])
        relative_module_path = Path(self.pointers[0]).expanduser().relative_to(source_dir)
        source_dir_name = Path(source_dir).name
        if self.compute.working_dir is not None:
            container_module_path = str(Path(self.compute.working_dir) / source_dir_name / relative_module_path)
        else:
            # Leave it as relative path
            container_module_path = str(Path(source_dir_name) / relative_module_path)
        self._remote_pointers = (
            container_module_path,
            self.pointers[1],
            self.pointers[2],
        )
        return self._remote_pointers

    @property
    def service_config(self) -> dict:
        """Knative service configuration loaded from Kubernetes API."""
        return self._service_config

    @service_config.setter
    def service_config(self, value: dict):
        self._service_config = value

    @property
    def base_endpoint(self):
        """Endpoint for the module."""
        if is_running_in_kubernetes():
            if not self._compute.endpoint:
                return self._compute._wait_for_endpoint()
            return self._compute.endpoint
        # URL format when using the NGINX proxy
        return f"http://localhost:{self._compute.client_port()}/{self.namespace}/{self.service_name}"

    @property
    def request_headers(self):
        if self.compute.freeze:
            return {}

        if self.deployment_timestamp:
            return {"X-Deployed-As-Of": self.deployment_timestamp}

        return {}

    @property
    def serialization(self):
        """Default serialization format for this module."""
        return self._serialization

    @serialization.setter
    def serialization(self, value: str):
        """Set the default serialization format for this module."""
        if value not in ["json", "pickle"]:
            raise ValueError("Serialization must be 'json' or 'pickle'")
        self._serialization = value

    @property
    def async_(self):
        """Whether to run the function or class methods in async mode."""
        return self._async

    @async_.setter
    def async_(self, value: bool):
        if not isinstance(value, bool):
            raise ValueError("`async_` must be a boolean")
        self._async = value

    @classmethod
    def from_name(
        cls,
        name: str,
        namespace: str = None,
        reload_prefixes: Union[str, List[str]] = [],
    ):
        """Reload an existing callable by its service name."""
        from kubernetes import client
        from kubernetes.config import ConfigException, load_incluster_config, load_kube_config

        import kubetorch as kt

        try:
            load_incluster_config()
        except ConfigException:
            load_kube_config()
        objects_api = client.CustomObjectsApi()
        apps_v1_api = client.AppsV1Api()
        core_v1_api = client.CoreV1Api()

        namespace = namespace or config.namespace
        if isinstance(reload_prefixes, str):
            reload_prefixes = [reload_prefixes]
        potential_names = get_names_for_reload_fallbacks(name=name, prefixes=reload_prefixes)

        # Use unified service discovery from BaseServiceManager
        from kubetorch.serving.service_manager import BaseServiceManager

        all_services = BaseServiceManager.discover_services_static(
            namespace=namespace, objects_api=objects_api, apps_v1_api=apps_v1_api
        )

        # Create name-to-service lookup for efficient searching
        service_dict = {svc["name"]: svc for svc in all_services}

        # Try to find the first matching service across all service types
        for candidate in potential_names:

            service_info = service_dict.get(candidate)
            if service_info is None:
                continue

            compute = kt.Compute.from_template(service_info)

            pods = core_v1_api.list_namespaced_pod(
                namespace=namespace,
                label_selector=f"kubetorch.com/service={name}",
            )
            volumes = []

            # TODO: handle case where service is scaled to 0?
            if pods.items:
                # Use runtime Pod spec
                pod = pods.items[0]
                for v in pod.spec.volumes or []:
                    if v.persistent_volume_claim:
                        existing_volume = kt.Volume.from_name(name=v.name)
                        volumes.append(existing_volume)

            module_args = compute.get_env_vars(
                [
                    "KT_FILE_PATH",
                    "KT_MODULE_NAME",
                    "KT_CLS_OR_FN_NAME",
                    "KT_CALLABLE_TYPE",
                    "KT_INIT_ARGS",
                ]
            )
            pointers = (
                module_args["KT_FILE_PATH"],
                module_args["KT_MODULE_NAME"],
                module_args["KT_CLS_OR_FN_NAME"],
            )

            if module_args.get("KT_CALLABLE_TYPE") == "cls":
                init_args = json.loads(module_args.get("KT_INIT_ARGS") or "{}")
                reloaded_module = kt.Cls(name=candidate, pointers=pointers, init_args=init_args)
            elif module_args.get("KT_CALLABLE_TYPE") == "fn":
                reloaded_module = kt.Fn(name=candidate, pointers=pointers)
            else:
                raise ValueError(f"Unknown module type: {module_args.get('KT_CALLABLE_TYPE')}")

            reloaded_module.service_name = candidate
            reloaded_module.compute = compute
            return reloaded_module

        raise ValueError(
            f"Service '{name}' not found in namespace '{namespace}' with reload_prefixes={reload_prefixes}"
        )

    def _client(self, *args, **kwargs):
        """Return the client through which to interact with the remote Module.
        If compute is not yet set, attempt to reload it.
        """
        if self._http_client is not None:
            return self._http_client

        if self.compute is None or self.service_config is None:
            namespace = self.namespace
            # When rebuilding the http client on reload, need to know whether to look for a prefix
            reload_prefixes = self.reload_prefixes
            logger.debug(
                f"Attempting to reload service '{self.service_name}' in namespace '{namespace}' with "
                f"reload_prefixes={reload_prefixes}"
            )
            reloaded_module = Module.from_name(
                name=self.service_name,
                namespace=namespace,
                reload_prefixes=reload_prefixes,
            )

            # Update settable attributes with reloaded module values
            self.compute = self.compute or reloaded_module.compute
            self.service_config = reloaded_module.service_config
            self.pointers = reloaded_module.pointers
            self.name = reloaded_module.name
            self.service_name = reloaded_module.service_name

        self._http_client = HTTPClient(
            base_url=self.endpoint(*args, **kwargs),
            compute=self.compute,
            service_name=self.service_name,
        )

        return self._http_client

    def endpoint(self, method_name: str = None):
        if not hasattr(self, "init_args"):
            return f"{self.base_endpoint}/{self.module_name}"
        else:
            return f"{self.base_endpoint}/{self.module_name}/{method_name}"

    def deploy(self):
        """
        Helper method to deploy modules specified by the @compute decorator. Used by `kt deploy` CLI command.
        Deploys the module to the specified compute.
        """
        if self.compute is None:
            raise ValueError("Compute must be set before deploying the module.")
        return self.to(self.compute, init_args=getattr(self, "init_args", None))

    async def deploy_async(self):
        """
        Async helper method to deploy modules specified by the @compute decorator. Used by `kt deploy` CLI command
        when multiple modules are present. Deploys the module to the specified compute asynchronously.
        """
        if self.compute is None:
            raise ValueError("Compute must be set before deploying the module.")
        return await self.to_async(self.compute, init_args=getattr(self, "init_args", None))

    def to(
        self,
        compute: "Compute",
        init_args: Dict = None,
        stream_logs: Union[bool, None] = None,
        verbosity: Union[LogVerbosity, str] = None,
        get_if_exists: bool = False,
        reload_prefixes: Union[str, List[str]] = [],
        dryrun: bool = False,
    ):
        """
        Send the function or class to the specified compute.

        Args:
            compute (Compute): The compute to send the function or class to.
            init_args (Dict, optional): Initialization arguments, which may be relevant for a class.
            stream_logs (bool, optional): Whether to stream logs during service launch. If None, uses the global
                config value.
            verbosity (Union[verbosity, str], optional): Verbosity of the logs streamed back to the client.
                If not specified, will stream select service logs. Can also be controlled globally via the config
                value `log_verbosity`. Supported values: "debug", "info", "critical".
            get_if_exists (Union[bool, List[str]], optional): Controls how service lookup is performed to determine
                whether to send the service to the compute.

                - If False (default): Do not attempt to reload the service.
                - If True: Attempt to find an existing service using a standard fallback order
                  (e.g., username, git branch, then prod). If found, re-use that existing service.
            reload_prefixes (Union[str, List[str]], optional): A list of prefixes to use when reloading the function
                (e.g., ["qa", "prod", "git-branch-name"]). If not provided, will use the current username,
                git branch, and prod.
            dryrun (bool, optional): Whether to setup and return the object as a dryrun (``True``),
                or to actually launch the compute and service (``False``).
        Returns:
            Module: The module instance.

        Example:

        .. code-block:: python

            import kubetorch as kt

            remote_cls = kt.cls(SlowNumpyArray, name=name).to(
                kt.Compute(cpus=".1"),
                init_args={"size": 10},
                stream_logs=True
            )
        """
        if not has_k8s_credentials():
            raise KubernetesCredentialsError(
                "Kubernetes credentials not found. Please ensure you are running in a Kubernetes cluster or have a valid kubeconfig file."
            )

        if get_if_exists:
            try:
                existing_service = self._get_existing_service(reload_prefixes)
                if existing_service:
                    logger.debug(f"Reusing existing service: {existing_service.service_name}")
                    return existing_service
            except Exception as e:
                logger.info(
                    f"Service {self.service_name} not found in namespace {self.compute.namespace} "
                    f"with reload_prefixes={reload_prefixes}: {str(e)}"
                )

        self.compute = compute
        self.compute.service_name = self.service_name

        if hasattr(self, "init_args"):
            self.init_args = init_args

        # We need the deployment timestamp at the start of the update so we know that artifacts deployed **after**
        # this time are part of the current deployment. We actually set it at the end to ensure that the deployment is
        # successful.
        logger.debug(f"Deploying module: {self.service_name}")
        deployment_timestamp = datetime.now(timezone.utc).isoformat()
        install_url, use_editable = get_kt_install_url(self.compute.freeze)

        if not dryrun and not self.compute.freeze:
            self._rsync_repo_and_image_patches(install_url, use_editable, init_args)

        self._launch_service(
            install_url,
            use_editable,
            init_args,
            deployment_timestamp,
            stream_logs,
            verbosity,
            dryrun,
        )

        return self

    async def to_async(
        self,
        compute: "Compute",
        init_args: Dict = None,
        stream_logs: Union[bool, None] = None,
        verbosity: Union[LogVerbosity, str] = None,
        get_if_exists: bool = False,
        reload_prefixes: Union[str, List[str]] = [],
        dryrun: bool = False,
    ):
        """
        Async version of the `.to` method. Send the function or class to the specified compute asynchronously.

        Args:
            compute (Compute): The compute to send the function or class to.
            init_args (Dict, optional): Initialization arguments, which may be relevant for a class.
            stream_logs (bool, optional): Whether to stream logs during service launch. If None, uses the global
                config value.
            verbosity (Union[verbosity, str], optional): Verbosity of the logs streamed back to the client.
                If not specified, will stream select service logs. Can also be controlled globally via the config
                value `log_verbosity`. Supported values: "debug", "info", "critical".
            get_if_exists (Union[bool, List[str]], optional): Controls how service lookup is performed to determine
                whether to send the service to the compute.

                - If False (default): Do not attempt to reload the service.
                - If True: Attempt to find an existing service using a standard fallback order
                  (e.g., username, git branch, then prod). If found, re-use that existing service.
            reload_prefixes (Union[str, List[str]], optional): A list of prefixes to use when reloading the function
                (e.g., ["qa", "prod", "git-branch-name"]). If not provided, will use the current username,
                git branch, and prod.
            dryrun (bool, optional): Whether to setup and return the object as a dryrun (``True``),
                or to actually launch the compute and service (``False``).
        Returns:
            Module: The module instance.

        Example:

        .. code-block:: python

            import kubetorch as kt

            remote_cls = await kt.cls(SlowNumpyArray, name=name).to_async(
                kt.Compute(cpus=".1"),
                init_args={"size": 10},
                stream_logs=True
            )
        """
        if get_if_exists:
            try:
                existing_service = await self._get_existing_service_async(reload_prefixes)
                if existing_service:
                    logger.debug(f"Reusing existing service: {existing_service.service_name}")
                    return existing_service
            except Exception as e:
                logger.info(
                    f"Service {self.compute.service_name} not found in namespace {self.compute.namespace} "
                    f"with reload_prefixes={reload_prefixes}: {str(e)}"
                )

        self.compute = compute
        self.compute.service_name = self.service_name

        if hasattr(self, "init_args"):
            self.init_args = init_args

        logger.debug(f"Deploying module: {self.service_name}")
        deployment_timestamp = datetime.now(timezone.utc).isoformat()
        install_url, use_editable = get_kt_install_url(self.compute.freeze)

        if not dryrun and not self.compute.freeze:
            await self._rsync_repo_and_image_patches_async(install_url, use_editable, init_args)

        await self._launch_service_async(
            install_url,
            use_editable,
            init_args,
            deployment_timestamp,
            stream_logs,
            verbosity,
            dryrun,
        )

        return self

    def _get_existing_service(self, reload_prefixes):
        try:
            existing_service = Module.from_name(
                self.service_name,
                namespace=self.namespace,
                reload_prefixes=reload_prefixes,
            )
            if existing_service:
                if self.compute:
                    # Replace the compute object, if the user has already constructed it locally
                    existing_service.compute = self.compute
                logger.info(
                    f"Existing service '{self.service_name}' found in namespace '{self.namespace}', not "
                    f"redeploying."
                )
                return existing_service
        except Exception as e:
            raise ValueError(
                f"Failed to reload service {self.service_name} in namespace {self.namespace} "
                f"and reload_prefixes={reload_prefixes}: {str(e)}"
            )

    async def _get_existing_service_async(self, reload_prefixes):
        try:
            existing_service = Module.from_name(
                self.service_name,
                namespace=self.namespace,
                reload_prefixes=reload_prefixes,
            )
            if existing_service:
                if self.compute:
                    # Replace the compute object, if the user has already constructed it locally
                    existing_service.compute = self.compute
                logger.info(
                    f"Existing service '{self.service_name}' found in namespace '{self.namespace}', not "
                    f"redeploying."
                )
                return existing_service
        except Exception as e:
            raise ValueError(
                f"Failed to reload service {self.service_name} in namespace {self.namespace} "
                f"and reload_prefixes={reload_prefixes}: {str(e)}"
            )

    def _get_rsync_dirs_and_dockerfile(self, install_url, use_editable, init_args):
        source_dir, has_kt_dir = locate_working_dir(self.pointers[0])
        rsync_dirs = [str(source_dir)]
        if not has_kt_dir:
            # Use the source file (.py) instead of directory
            source_file = Path(f"{self.pointers[0]}/{self.pointers[1]}.py")
            rsync_dirs = [str(source_file)]
            logger.info(f"Package root not found; syncing file {source_file}")
        else:
            logger.info(f"Package root identified at {source_dir}; syncing directory")

        if install_url.endswith(".whl") or (use_editable and install_url != str(source_dir)):
            rsync_dirs.append(install_url)

        pointer_env_vars = self._get_pointer_env_vars(self.remote_pointers)
        metadata_env_vars = self._get_metadata_env_vars(init_args)
        service_dockerfile = self._get_service_dockerfile({**pointer_env_vars, **metadata_env_vars})
        return rsync_dirs, service_dockerfile

    def _rsync_repo_and_image_patches(self, install_url, use_editable, init_args):
        logger.debug("Rsyncing data to the rsync pod")
        rsync_dirs, service_dockerfile = self._get_rsync_dirs_and_dockerfile(install_url, use_editable, init_args)
        self._construct_and_rsync_files(rsync_dirs, service_dockerfile)
        logger.debug(f"Rsync completed for service {self.service_name}")

    async def _rsync_repo_and_image_patches_async(self, install_url, use_editable, init_args):
        logger.debug("Rsyncing data to the rsync pod")
        rsync_dirs, service_dockerfile = self._get_rsync_dirs_and_dockerfile(install_url, use_editable, init_args)
        await self._construct_and_rsync_files_async(rsync_dirs, service_dockerfile)
        logger.debug(f"Rsync completed for service {self.service_name}")

    def _launch_service(
        self,
        install_url,
        use_editable,
        init_args,
        deployment_timestamp,
        stream_logs,
        verbosity,
        dryrun,
    ):
        # Start log streaming if enabled
        stop_event = threading.Event()
        log_thread = None
        if stream_logs is None:
            stream_logs = config.stream_logs or False

        launch_request_id = "-"
        if stream_logs and not dryrun:
            if verbosity is None:
                verbosity = config.log_verbosity

            # Create a unique request ID for this launch sequence
            launch_request_id = f"launch_{generate_unique_request_id('launch', deployment_timestamp)}"

            # Start log streaming in a separate thread
            log_thread = threading.Thread(
                target=self._stream_launch_logs,
                args=(
                    launch_request_id,
                    stop_event,
                    verbosity,
                    deployment_timestamp,
                ),
            )
            log_thread.daemon = True
            log_thread.start()

        try:
            startup_rsync_command = self._startup_rsync_command(use_editable, install_url, dryrun)

            # Launch the compute in the form of a service with the requested resources
            service_config = self.compute._launch(
                service_name=self.compute.service_name,
                install_url=install_url if not use_editable else None,
                pointer_env_vars=self._get_pointer_env_vars(self.remote_pointers),
                metadata_env_vars=self._get_metadata_env_vars(init_args),
                startup_rsync_command=startup_rsync_command,
                launch_id=launch_request_id,
                dryrun=dryrun,
            )
            self.service_config = service_config

            if not self.compute.freeze and not dryrun:
                self.deployment_timestamp = self.compute.service_manager.update_deployment_timestamp_annotation(
                    service_name=self.service_name,
                    new_timestamp=deployment_timestamp,
                )
            if not dryrun:
                self.compute._check_service_ready()
                # Additional health check to ensure HTTP server is ready
                self._wait_for_http_health()
        finally:
            # Stop log streaming
            if log_thread:
                stop_event.set()

    async def _launch_service_async(
        self,
        install_url,
        use_editable,
        init_args,
        deployment_timestamp,
        stream_logs,
        verbosity,
        dryrun,
    ):
        # Start log streaming if enabled
        stop_event = asyncio.Event()
        log_task = None
        if stream_logs is None:
            stream_logs = config.stream_logs or False

        launch_request_id = "-"
        if stream_logs and not dryrun:
            if verbosity is None:
                verbosity = config.log_verbosity

            # Create a unique request ID for this launch sequence
            launch_request_id = f"launch_{generate_unique_request_id('launch', deployment_timestamp)}"

            # Start log streaming as an async task
            log_task = asyncio.create_task(
                self._stream_launch_logs_async(
                    launch_request_id,
                    stop_event,
                    verbosity,
                    deployment_timestamp,
                )
            )

        try:
            startup_rsync_command = self._startup_rsync_command(use_editable, install_url, dryrun)

            # Launch the compute in the form of a service with the requested resources
            # Use the async version of _launch
            service_config = await self.compute._launch_async(
                service_name=self.compute.service_name,
                install_url=install_url if not use_editable else None,
                pointer_env_vars=self._get_pointer_env_vars(self.remote_pointers),
                metadata_env_vars=self._get_metadata_env_vars(init_args),
                startup_rsync_command=startup_rsync_command,
                launch_id=launch_request_id,
                dryrun=dryrun,
            )
            self.service_config = service_config

            if not self.compute.freeze and not dryrun:
                self.deployment_timestamp = self.compute.service_manager.update_deployment_timestamp_annotation(
                    service_name=self.service_name,
                    new_timestamp=deployment_timestamp,
                )
            if not dryrun:
                await self.compute._check_service_ready_async()
                await self._wait_for_http_health_async()
        finally:
            # Stop log streaming
            if log_task:
                stop_event.set()
                try:
                    await asyncio.wait_for(log_task, timeout=2.0)
                except asyncio.TimeoutError:
                    log_task.cancel()
                    try:
                        await log_task
                    except asyncio.CancelledError:
                        pass

    def _get_service_dockerfile(self, metadata_env_vars):
        image_instructions = self.compute._image_setup_and_instructions()

        if image_instructions:
            image_instructions += "\n"
        for key, val in metadata_env_vars.items():
            if isinstance(val, Dict):
                val = json.dumps(val)
            image_instructions += f"ENV {key} {val}\n"

        logger.debug(f"Generated Dockerfile for service {self.service_name}:\n{image_instructions}")
        return image_instructions

    def _construct_and_rsync_files(self, rsync_dirs, service_dockerfile):
        with tempfile.TemporaryDirectory() as tmpdir:
            temp_file = Path(tmpdir) / ".kt" / "image.dockerfile"
            temp_file.parent.mkdir(parents=True, exist_ok=True)
            temp_file.write_text(service_dockerfile)

            source_dir = str(Path(tmpdir) / ".kt")
            rsync_dirs.append(source_dir)

            logger.debug(f"Rsyncing directories: {rsync_dirs}")
            if is_running_in_kubernetes():
                self.compute.rsync_in_cluster(rsync_dirs)
            else:
                self.compute.rsync(rsync_dirs)

    async def _construct_and_rsync_files_async(self, rsync_dirs, service_dockerfile):
        with tempfile.TemporaryDirectory() as tmpdir:
            temp_file = Path(tmpdir) / ".kt" / "image.dockerfile"
            temp_file.parent.mkdir(parents=True, exist_ok=True)
            temp_file.write_text(service_dockerfile)

            source_dir = str(Path(tmpdir) / ".kt")
            rsync_dirs.append(source_dir)

            logger.debug(f"Rsyncing directories: {rsync_dirs}")
            if is_running_in_kubernetes():
                await self.compute.rsync_in_cluster_async(rsync_dirs)
            else:
                await self.compute.rsync_async(rsync_dirs)

    def _startup_rsync_command(self, use_editable, install_url, dryrun):
        if dryrun:
            return None

        if use_editable or (install_url and install_url.endswith(".whl")):
            # rsync from the rsync pod's file system directly
            startup_cmd = self.compute._rsync_svc_url()
            cmd = f"rsync -av {startup_cmd} ."
            return cmd

        return None

    def teardown(self):
        """Delete the service and all associated resources."""
        logger.info(f"Deleting service: {self.service_name}")

        # Use the compute's service manager - it already knows the correct type!
        teardown_success = self.compute.service_manager.teardown_service(
            service_name=self.service_name,
        )

        if not teardown_success:
            logger.error(f"Failed to teardown service {self.service_name}")
            return

        configmaps = load_configmaps(
            core_api=self.compute.core_api,
            service_name=self.service_name,
            namespace=self.compute.namespace,
        )
        if configmaps:
            logger.info(f"Deleting {len(configmaps)} configmap{'' if len(configmaps) == 1 else 's'}")
            delete_configmaps(
                core_api=self.compute.core_api,
                configmaps=configmaps,
                namespace=self.compute.namespace,
            )

        logger.info("Deleting service data from cache in rsync pod")
        delete_cached_service_data(
            core_api=self.compute.core_api,
            service_name=self.service_name,
            namespace=self.compute.namespace,
        )

    def _get_pointer_env_vars(self, remote_pointers):
        (container_file_path, module_name, cls_or_fn_name) = remote_pointers
        return {
            "KT_FILE_PATH": container_file_path,
            "KT_MODULE_NAME": module_name,
            "KT_CLS_OR_FN_NAME": cls_or_fn_name,
        }

    def _get_metadata_env_vars(
        self,
        init_args: Dict,
    ) -> Dict:
        # TODO: add other callable metadata in addition to pointers (`is_generator`, `is_async`, etc.)
        import json

        distributed_config = self.compute.distributed_config
        return {
            "KT_INIT_ARGS": init_args,
            "KT_CALLABLE_TYPE": self.MODULE_TYPE,
            "KT_DISTRIBUTED_CONFIG": json.dumps(distributed_config) if distributed_config else None,
        }

    def _stream_launch_logs(
        self,
        request_id: str,
        stop_event: threading.Event,
        verbosity: LogVerbosity,
        deployment_timestamp: str,
    ):
        """Stream logs and events during service launch sequence."""
        try:
            # Only use "kubetorch" container to exclude queue-proxy (e.g. Knative sidecars) container logs which
            # are spammy with tons of healthcheck calls
            pod_query = f'{{k8s_container_name="kubetorch"}} | json | request_id="{request_id}"'
            event_query = f'{{service_name="unknown_service"}} | json | k8s_object_name=~"{self.service_name}.*" | k8s_namespace_name="{self.namespace}"'

            encoded_pod_query = urllib.parse.quote_plus(pod_query)
            encoded_event_query = urllib.parse.quote_plus(event_query)
            logger.debug(f"Streaming launch logs and events for service {self.service_name}")

            def start_log_threads(host, port):
                def run_pod_logs():
                    self._run_log_stream(
                        request_id,
                        stop_event,
                        host,
                        port,
                        encoded_pod_query,
                        verbosity,
                        deployment_timestamp,
                        dedup=True,
                    )

                def run_event_logs():
                    self._run_log_stream(
                        request_id,
                        stop_event,
                        host,
                        port,
                        encoded_event_query,
                        verbosity,
                        deployment_timestamp,
                    )

                pod_thread = threading.Thread(target=run_pod_logs, daemon=True)
                event_thread = threading.Thread(target=run_event_logs, daemon=True)

                pod_thread.start()
                event_thread.start()

                # Don't block indefinitely on joins - use short timeouts
                pod_thread.join(timeout=1.0)
                event_thread.join(timeout=1.0)

            base_url = service_url()
            host, port = extract_host_port(base_url)
            logger.debug(f"Streaming launch logs with url={base_url} host={host} and local port {port}")
            start_log_threads(host, port)

        except Exception as e:
            logger.error(f"Failed to stream launch logs: {e}")
            raise e

    async def _stream_launch_logs_async(
        self,
        request_id: str,
        stop_event: asyncio.Event,
        verbosity: LogVerbosity,
        deployment_timestamp: str,
    ):
        """Async version of _stream_launch_logs. Stream logs and events during service launch sequence."""
        try:
            # Only use "kubetorch" container to exclude queue-proxy (e.g. Knative sidecars) container logs which
            # are spammy with tons of healthcheck calls
            pod_query = f'{{k8s_container_name="kubetorch"}} | json | request_id="{request_id}"'
            event_query = f'{{service_name="unknown_service"}} | json | k8s_object_name=~"{self.service_name}.*" | k8s_namespace_name="{self.namespace}"'

            encoded_pod_query = urllib.parse.quote_plus(pod_query)
            encoded_event_query = urllib.parse.quote_plus(event_query)
            logger.debug(f"Streaming launch logs and events for service {self.service_name}")

            base_url = await service_url_async()
            host, port = extract_host_port(base_url)
            logger.debug(f"Streaming launch logs with url={base_url} host={host} and local port {port}")

            # Create async tasks for both log streams
            pod_task = asyncio.create_task(
                self._stream_logs_websocket(
                    request_id,
                    stop_event,
                    host=host,
                    port=port,
                    query=encoded_pod_query,
                    log_verbosity=verbosity,
                    deployment_timestamp=deployment_timestamp,
                    dedup=True,
                )
            )

            event_task = asyncio.create_task(
                self._stream_logs_websocket(
                    request_id,
                    stop_event,
                    host=host,
                    port=port,
                    query=encoded_event_query,
                    log_verbosity=verbosity,
                    deployment_timestamp=deployment_timestamp,
                )
            )

            # Wait for both tasks to complete or be cancelled
            try:
                await asyncio.gather(pod_task, event_task, return_exceptions=True)
            except Exception as e:
                logger.error(f"Error in async log streaming: {e}")

        except Exception as e:
            logger.error(f"Failed to stream launch logs: {e}")
            raise e

    def _run_log_stream(
        self,
        request_id: str,
        stop_event: threading.Event,
        host: str,
        port: int,
        query: str,
        log_verbosity: LogVerbosity,
        deployment_timestamp: str,
        dedup: bool = False,
    ):
        """Helper to run log streaming in an event loop"""
        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)
        try:
            loop.run_until_complete(
                self._stream_logs_websocket(
                    request_id,
                    stop_event,
                    host=host,
                    port=port,
                    query=query,
                    log_verbosity=log_verbosity,
                    deployment_timestamp=deployment_timestamp,
                    dedup=dedup,
                )
            )
        finally:
            loop.close()

    async def _run_log_stream_async(
        self,
        request_id: str,
        stop_event: asyncio.Event,
        host: str,
        port: int,
        query: str,
        log_verbosity: LogVerbosity,
        deployment_timestamp: str,
        dedup: bool = False,
    ):
        """Async helper to run log streaming directly in the current event loop"""
        await self._stream_logs_websocket(
            request_id,
            stop_event,
            host=host,
            port=port,
            query=query,
            log_verbosity=log_verbosity,
            deployment_timestamp=deployment_timestamp,
            dedup=dedup,
        )

    async def _stream_logs_websocket(
        self,
        request_id: str,
        stop_event: Union[threading.Event, asyncio.Event],
        host: str,
        port: int,
        query: str,
        log_verbosity: LogVerbosity,
        deployment_timestamp: str,
        dedup: bool = False,
    ):
        """Stream logs and events using Loki's websocket tail endpoint"""
        try:
            uri = f"ws://{host}:{port}/loki/api/v1/tail?query={query}"

            # Track the last timestamp we've seen to avoid duplicates
            last_timestamp = None

            # Track when we should stop
            stop_time = None

            # Track most recent deployment timestamp to filter out old logs / events
            start_timestamp = iso_timestamp_to_nanoseconds(deployment_timestamp)

            shown_event_messages = set()

            # Track seen log messages for deduplication
            seen_log_messages = set() if dedup else None

            # For formatting the server setup logs
            formatters = {}
            base_formatter = ServerLogsFormatter()
            websocket = None
            try:
                # Add timeout to prevent hanging connections
                websocket = await websockets.connect(
                    uri,
                    close_timeout=10,  # Max time to wait for close handshake
                    ping_interval=20,  # Send ping every 20 seconds
                    ping_timeout=10,  # Wait 10 seconds for pong
                )
                while True:
                    # If stop event is set, start counting down
                    # Handle both threading.Event and asyncio.Event
                    is_stop_set = stop_event.is_set() if hasattr(stop_event, "is_set") else stop_event.is_set()
                    if is_stop_set and stop_time is None:
                        stop_time = time.time() + 2  # 2 second grace period

                    # If we're past the grace period, exit
                    if stop_time is not None and time.time() > stop_time:
                        break

                    try:
                        # Use shorter timeout during grace period
                        timeout = 0.1 if stop_time is not None else 1.0
                        message = await asyncio.wait_for(websocket.recv(), timeout=timeout)
                        data = json.loads(message)

                        if data.get("streams"):
                            for stream in data["streams"]:
                                labels = stream.get("stream", {})
                                is_event = "k8s_event_count" in list(labels.keys())
                                for value in stream["values"]:
                                    ts_ns = int(value[0])
                                    if start_timestamp is not None and ts_ns < start_timestamp:
                                        continue
                                    log_line = value[1]
                                    if is_event:
                                        event_type = labels.get("detected_level", "")
                                        if log_verbosity == LogVerbosity.CRITICAL and event_type == "Normal":
                                            # skip Normal events in MINIMAL
                                            continue

                                        try:
                                            msg = log_line
                                            reason = (labels.get("k8s_event_reason", ""),)

                                            # Note: relevant starting in release 0.1.19 (using OTel instead of Alloy)
                                            if isinstance(reason, tuple):
                                                reason = reason[0]

                                            event_type = labels.get("detected_level", "")

                                            if reason == "Unhealthy" and (
                                                "HTTP probe failed with statuscode: 503" in msg
                                                or "Startup probe failed" in msg
                                            ):
                                                # HTTP probe failures are expected during setup
                                                continue

                                            ignore_patterns = (
                                                "queue-proxy",
                                                "resolving reference: address not set for kind = service",
                                                "failed to get private k8s service endpoints:",
                                            )
                                            # Ignore queue-proxy events and gateway setup events
                                            if any(pattern in msg.lower() for pattern in ignore_patterns):
                                                continue

                                            if msg in shown_event_messages:
                                                # Only show unique event messages
                                                continue

                                            shown_event_messages.add(msg)

                                        except Exception:
                                            # If parsing fails, just print the event as is
                                            pass

                                        if event_type == "Normal":
                                            if log_verbosity in [
                                                LogVerbosity.INFO,
                                                LogVerbosity.DEBUG,
                                            ]:
                                                print(f'[EVENT] reason={reason} "{msg}"')
                                        else:
                                            print(f'[EVENT] type={event_type} reason={reason} "{msg}"')
                                        continue

                                    # Skip if we've already seen this timestamp
                                    if last_timestamp is not None and value[0] <= last_timestamp:
                                        continue
                                    last_timestamp = value[0]
                                    if log_verbosity in [
                                        LogVerbosity.DEBUG,
                                        LogVerbosity.INFO,
                                    ]:
                                        try:
                                            log_dict = json.loads(log_line)
                                        except json.JSONDecodeError:
                                            # setup steps pre server start are not JSON formatted
                                            log_dict = None

                                        if log_dict is not None:
                                            # at this stage we are post setup
                                            pod_name = log_dict.get("pod", request_id)
                                            levelname = log_dict.get("levelname", "INFO")
                                            ts = log_dict.get("asctime")
                                            message = log_dict.get("message", "")

                                            if (
                                                log_verbosity == LogVerbosity.CRITICAL
                                                and levelname not in ["ERROR", "CRITICAL"]
                                            ) or (log_verbosity == LogVerbosity.INFO and levelname == "DEBUG"):
                                                continue

                                            log_line = f"{levelname} | {ts} | {message}"
                                            if pod_name not in formatters:
                                                formatters[pod_name] = ServerLogsFormatter(pod_name)
                                            formatter = formatters[pod_name]
                                        else:
                                            # streaming pre server setup logs, before we have the pod name
                                            formatter = base_formatter

                                        newline = "" if log_dict is None else None
                                        formatted_line = f"{formatter.start_color}{f'({self.service_name}) '}{log_line}{formatter.reset_color}"

                                        # Check for duplicates if dedup is enabled
                                        if seen_log_messages is not None:
                                            if message in seen_log_messages:
                                                continue
                                            seen_log_messages.add(message)

                                        print(formatted_line, end=newline)
                    except asyncio.TimeoutError:
                        # Timeout is expected, just continue the loop
                        continue
                    except websockets.exceptions.ConnectionClosed as e:
                        logger.debug(f"WebSocket connection closed: {str(e)}")
                        break
            finally:
                if websocket:
                    try:
                        # Use wait_for to prevent hanging on close
                        await asyncio.wait_for(websocket.close(), timeout=1.0)
                    except (asyncio.TimeoutError, Exception):
                        pass
        except Exception as e:
            logger.error(f"Error in websocket stream: {e}")
            raise e
        finally:
            # Ensure websocket is closed even if we didn't enter the try block
            if websocket:
                try:
                    # Use wait_for to prevent hanging on close
                    await asyncio.wait_for(websocket.close(), timeout=1.0)
                except (asyncio.TimeoutError, Exception):
                    pass

    def _wait_for_http_health(self, timeout=60, retry_interval=0.1, backoff=2, max_interval=10):
        """Wait for the HTTP server to be ready by checking the /health endpoint.

        Args:
            timeout: Maximum time to wait in seconds
            retry_interval: Time between health check attempts in seconds
        """
        import time

        logger.info(f"Waiting for HTTP server to be ready for service {self.service_name}")
        start_time = time.time()

        while time.time() - start_time < timeout:
            try:
                client = self._client()
                response = client.get(
                    endpoint=f"{self.base_endpoint}/health",
                    headers=self.request_headers,
                )
                if response.status_code == 200:
                    logger.info(f"HTTP server is ready for service {self.service_name}")
                    return
                else:
                    logger.debug(f"Health check returned status {response.status_code}, retrying...")

            except VersionMismatchError as e:
                raise e

            except Exception as e:
                logger.debug(f"Health check failed: {e}, retrying...")

            time.sleep(retry_interval)
            retry_interval *= backoff  # Exponential backoff
            # Cap the retry interval to a maximum value
            retry_interval = min(retry_interval, max_interval)

        # If we get here, we've timed out
        logger.warning(f"HTTP health check timed out after {timeout}s for service {self.service_name}")

    async def _wait_for_http_health_async(self, timeout=60, retry_interval=0.1, backoff=2, max_interval=10):
        """Async version of _wait_for_http_health. Wait for the HTTP server to be ready by checking the /health endpoint.

        Args:
            timeout: Maximum time to wait in seconds
            retry_interval: Time between health check attempts in seconds
        """
        import asyncio

        logger.info(f"Waiting for HTTP server to be ready for service {self.service_name}")
        start_time = time.time()

        while time.time() - start_time < timeout:
            try:
                client = self._client()
                response = client.get(
                    endpoint=f"{self.base_endpoint}/health",
                    headers=self.request_headers,
                )
                if response.status_code == 200:
                    logger.info(f"HTTP server is ready for service {self.service_name}")
                    return
                else:
                    logger.debug(f"Health check returned status {response.status_code}, retrying...")
            except Exception as e:
                logger.debug(f"Health check failed: {e}, retrying...")

            await asyncio.sleep(retry_interval)
            retry_interval *= backoff  # Exponential backoff
            # Cap the retry interval to a maximum value
            retry_interval = min(retry_interval, max_interval)

        # If we get here, we've timed out
        logger.warning(f"HTTP health check timed out after {timeout}s for service {self.service_name}")

    def __getstate__(self):
        """Remove local stateful values before pickle serialization."""
        state = self.__dict__.copy()
        # Remove local stateful values that shouldn't be serialized
        state["_http_client"] = None
        state["_service_config"] = None
        state["_remote_pointers"] = None
        # Pointers need to be converted to not be absolute paths if we're passing
        # the service elsewhere, e.g. into another service
        state["pointers"] = self.remote_pointers
        return state

    def __setstate__(self, state):
        """Restore state after pickle deserialization."""
        self.__dict__.update(state)
        # Reset local stateful values to None to ensure clean initialization
        self._http_client = None
        self._service_config = None
        self._remote_pointers = None

    def __del__(self):
        if hasattr(self, "_http_client") and self._http_client is not None:
            try:
                self._http_client.close()
            except Exception as e:
                logger.debug(f"Error closing HTTPClient in Module deletion: {e}")
            finally:
                self._http_client = None
